/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

// This module requires Eigen library
#ifdef HAVE_EIGEN3

#include "../io/encoding.hpp"
#include <Eigen/Dense>
#include <stdint.h>

namespace broccoli {
namespace curve {
    //! Abstract representation of a quaternion curve (interpolation is not specified yet)
    /*!
     * Class acts as a prototype and declares all functions and members a quaternion curve should provide.
     *
     * The underlying curve is given with respect to `position` which is **not** the time in general. Instead
     * `position` describes the interpolation parameter within the interval \f$ [0, 1] \f$ where 0 corresponds to the
     * start-quaternion and 1 to the end-quaternion. This implies, that one has to be careful when using the
     * \ref evaluate function to get derivatives (use chain rule!).
     */
    class QuaternionCurve {
    public:
        //! Specification of possible interpolation types
        enum class InterpolationType : uint8_t {
            SLERP = 0, //!< <b>S</b>pherical <b>L</b>inear int<b>ERP</b>olation between start-quaternion and end-quaternion
            BEZIER_CUBIC, //!< Interpolation using cubic bezier curves (two extra quaternions as control points)
            INTERPOLATIONTYPE_COUNT //!< Total count of elements
        };

        //! Returns the string representation of the given interpolation type
        static std::string interpolationTypeString(const InterpolationType& type)
        {
            // Check type
            switch (type) {
            case InterpolationType::SLERP:
                return "SLERP";
            case InterpolationType::BEZIER_CUBIC:
                return "BEZIER_CUBIC";
            default:
                return "UNKNOWN";
            }
        }

        //! Constructor
        QuaternionCurve()
        {
        }

        //! Destructor
        virtual ~QuaternionCurve()
        {
        }

        //! Comparison operator: **equality**
        virtual bool operator==(const QuaternionCurve& reference) const
        {
            (void)reference; // Unused

            // No own members -> always equal
            return true;
        }

        //! Comparison operator: **inequality**
        virtual bool operator!=(const QuaternionCurve& reference) const { return !(*this == reference); }

        //! Evaluation of (derivative of) curve
        /*!
         * \warning For \p derivationOrder > 0 the evaluate function does **not** return the angular velocity/acceleration/...
         * but rather the "pure" derivative of the quaternion!
         *
         * \param [in] position Position to evaluate (derivative) at (has to be within \f$ [0, 1] \f$)
         * \param [in] derivationOrder Order of the derivation (use 0 for base function, i.e. no derivation)
         * \return Value or derivative at specified position
         */
        virtual Eigen::Quaterniond evaluate(const double& position, const unsigned int& derivationOrder = 0) const = 0;

        // Logging
        // -------
        //! Encode member data as XML element and add it to the specified stream
        /*!
         * \param [in,out] stream The stream to which the member data should be appended to.
         * \param [in] XMLIndentationLevel Level of indentation in XML format (used to add spaces).
         * \param [in] XMLTabWhiteSpaces Count of whitespaces corresponding to a tab.
         * \param [in] numericFormat C-format specifier for numeric values.
         * \return Count of appended characters (elements in stream).
         */
        virtual io::encoding::CharacterStreamSize encodeToXML(io::encoding::CharacterStream& stream, const size_t& XMLIndentationLevel, const size_t& XMLTabWhiteSpaces, const std::string& numericFormat = "%.7g") const = 0;

    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW // <-- Proper 128 bit alignment of member data necessary for Eigen vectorization
    };
} // namespace curve
} // namespace broccoli

#endif // HAVE_EIGEN3
