/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

// This module requires Eigen library
#ifdef HAVE_EIGEN3

#include "../core/math.hpp"
#include "../io/encoding.hpp"
#include "QuaternionCurve.hpp"
#include "QuaternionSLERPCurve.hpp"
#include <Eigen/Dense>
#include <array>
#include <math.h>
#include <vector>

namespace broccoli {
namespace curve {
    // Configuration
    const double minimumQuaternionSplineSegmentProportion = 1e-6; //!< Minimum proportion of quaternion spline segments (used to avoid numeric errors)

    //! Specification of quaternion spline interpolation methods
    enum class QuaternionSplineInterpolationMethod : uint8_t {
        SLERP = 0, //!< <b>S</b>pherical <b>L</b>inear int<b>ERP</b>olation
        BEZIER_CUBIC_CUSTOM, //!< Quaternion interpolation using cubic bezier curves (**custom choice** of bezier *control-points*)
        BEZIER_CUBIC_SMOOTH, //!< Quaternion interpolation using cubic bezier curves (**automatic choice** of bezier *control-points* for a smooth animation)
        QUATERNIONSPLINEINTERPOLATIONMETHOD_COUNT //!< Total count of elements
    };

    //! Returns the string representation of the given quaternion spline interpolation method
    static inline std::string quaternionSplineInterpolationMethodString(const QuaternionSplineInterpolationMethod& method)
    {
        // Check method
        switch (method) {
        case QuaternionSplineInterpolationMethod::SLERP:
            return "SLERP";
        case QuaternionSplineInterpolationMethod::BEZIER_CUBIC_CUSTOM:
            return "BEZIER_CUBIC_CUSTOM";
        case QuaternionSplineInterpolationMethod::BEZIER_CUBIC_SMOOTH:
            return "BEZIER_CUBIC_SMOOTH";
        default:
            return "UNKNOWN";
        }
    }

    //! Specification of result types for quaternion spline algorithms
    enum class QuaternionSplineResult : uint8_t {
        UNKNOWN = 0, //!< Unknown result (this should **never** happen)
        SUCCESS, //!< Algorithm was successful
        ERROR_INVALID_SPLINE_NO_SEGMENTS, //!< An **error** occured: spline is invalid (it has to contain at least one segment)
        ERROR_INVALID_SPLINE_PROPORTION_COUNT, //!< An **error** occured: spline is invalid (proportion count does not match segment count)
        ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE, //!< An **error** occured: spline is invalid (negative proportion found)
        ERROR_INVALID_SPLINE_PROPORTION_SUM, //!< An **error** occured: spline is invalid (sum of proportions does not match 1)
        ERROR_INVALID_PARAMETERS, //!< An **error** occured: invalid input parameters
        ERROR_NOT_IMPLEMENTED, //!< An **error** occured: the desired algorithm is not yet implemented
        QUATERNIONSPLINERESULT_COUNT //!< Total count of elements
    };

    //! Returns the string representation of the given result type for quaternion spline algorithms
    static inline std::string quaternionSplineResultString(const QuaternionSplineResult& result)
    {
        // Check result
        switch (result) {
        case QuaternionSplineResult::UNKNOWN:
            return "UNKNOWN";
        case QuaternionSplineResult::SUCCESS:
            return "SUCCESS";
        case QuaternionSplineResult::ERROR_INVALID_SPLINE_NO_SEGMENTS:
            return "ERROR_INVALID_SPLINE_NO_SEGMENTS";
        case QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT:
            return "ERROR_INVALID_SPLINE_PROPORTION_COUNT";
        case QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE:
            return "ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE";
        case QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM:
            return "ERROR_INVALID_SPLINE_PROPORTION_SUM";
        case QuaternionSplineResult::ERROR_INVALID_PARAMETERS:
            return "ERROR_INVALID_PARAMETERS";
        case QuaternionSplineResult::ERROR_NOT_IMPLEMENTED:
            return "ERROR_NOT_IMPLEMENTED";
        default:
            return "UNKNOWN";
        }
    }

    //! Template class for quaternion splines as concatenation of segments (quaternion curves)
    /*!
     * \warning This class is restricted to **unit quaternions**!
     *
     * Contains an ordered list of quaternion spline segments (\ref m_segments). The \p CurveType has to be specified at compile time. Each segment has an underlying
     * curve, which is bounded by the *segment-position* \f$ \in [0,\,1] \f$. The complete spline is also bounded by the *spline-position \f$ \in [0,\, 1] \f$
     * where in general the *spline-position* is not the *segment-position*. Thus there has to be a mapping between *spline-position* and *segment-position*
     * such that we can find out which segment corresponds to which *spline-position*. This is done on a percentage basis (\ref m_segmentProportions).
     * Considering the derivative of the underlying curve, the gradient relates to the *spline-coordinates*, **not** to the segment coordinates
     * (mapping is performed automatically).
     *
     * \par Segmentation
     * \code
     * |0--------1|0--------1|0-------1|  <-- segment-coordinates
     *   Segment0 | Segment1 | Segment2
     *   Propor.0 | Propor.1 | Propor.2
     *
     * |0-----------------------------1|  <-- spline-coordinates
     *              Spline
     * \endcode
     */
    template <class CurveType>
    class QuaternionSpline {
    public:
        //! Constructor
        QuaternionSpline()
        {
        }

        //! Destructor
        virtual ~QuaternionSpline()
        {
        }

        //! Comparison operator: **equality**
        virtual bool operator==(const QuaternionSpline<CurveType>& reference) const
        {
            // Compare segments
            if (m_segments != reference.m_segments)
                return false;

            // Compare segment proportions
            if (m_segmentProportions != reference.m_segmentProportions)
                return false;

            // Otherwise -> equality
            return true;
        }

        //! Comparison operator: **inequality**
        virtual bool operator!=(const QuaternionSpline<CurveType>& reference) const { return !(*this == reference); }

        // Members
        std::vector<CurveType, Eigen::aligned_allocator<CurveType>> m_segments; //!< Ordered list of spline segments (each segment has its own boundaries: *segment-position* \f$ \in [0,\,1] \f$)
        std::vector<double> m_segmentProportions; //!< Proportion of single segments relative to complete spline "duration" (each element has to be \f$ >=0 \f$ and the sum of all elements has to be exactly 1)

        //! Checks, if the spline is properly defined (proper definition of segments and proportions)
        /*!
         * \param [out] result Pointer to flag in which the result should be stored
         * \return `true`, if the spline is valid, `false` otherwise.
         */
        inline bool isValid(QuaternionSplineResult* const result = nullptr) const
        {
            // Check, if there is at least one segment
            if (m_segments.size() == 0) {
                if (result != nullptr)
                    *result = QuaternionSplineResult::ERROR_INVALID_SPLINE_NO_SEGMENTS;
                return false;
            }

            // Check, if segment and proportion count are the same
            if (m_segments.size() != m_segmentProportions.size()) {
                if (result != nullptr)
                    *result = QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT;
                return false;
            }

            // Check, proportion entries
            // (each entry has to be >=0 and the sum of all entries has to be 1.0)
            double sumOfProportions = 0;
            for (size_t i = 0; i < m_segmentProportions.size(); i++) {
                if (m_segmentProportions[i] < 0) {
                    if (result != nullptr)
                        *result = QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE;
                    return false;
                }
                sumOfProportions += m_segmentProportions[i];
            }
            if (fabs(sumOfProportions - 1.0) > minimumQuaternionSplineSegmentProportion) {
                if (result != nullptr)
                    *result = QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM;
                return false;
            }

            // Otherwise -> valid
            if (result != nullptr)
                *result = QuaternionSplineResult::SUCCESS;
            return true;
        }

        //! Computes the segment boundaries (control points positions) represented in **spline coordinates** ([0, 1])
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \return Ordered list of segment boundaries, or empty list in case of an error.
         *
         * \code
         * |--------S--------| <-- spline
         * |---s1---|---s2---| <-- spline segments
         * B1       B2      B3 <-- segment boundaries
         * 0       0.x       1 <-- returned values
         * \endcode
         */
        inline std::vector<double> getSegmentBoundaries() const
        {
            // Initialize helpers
            std::vector<double> boundaries;

            // Abort, if there are no segments
            if (m_segmentProportions.size() == 0)
                return boundaries;

            // Pre-allocate memory for speedup
            boundaries.reserve(m_segmentProportions.size() + 1);

            // Add "left" boundary of fist segment
            boundaries.push_back(0);

            // Get "right" boundary of all segments
            for (size_t i = 0; i < m_segmentProportions.size(); i++)
                boundaries.push_back(boundaries.back() + m_segmentProportions[i]);

            // Force last boundary position to match exactly 1 (to avoid numeric errors)
            boundaries.back() = 1;

            // Pass back list
            return boundaries;
        }

        //! Removes segments with zero proportion (i.e. no contribution to the spline)
        /*! \return Count of removed segments */
        inline size_t removeEmptySegments()
        {
            // Abort, if there are no segments (or invalid dimensions)
            if (m_segments.size() == 0 || m_segments.size() != m_segmentProportions.size())
                return 0;

            // Iterate through segments
            size_t removedSegments = 0;
            size_t nextFreeSlot = 0; // Next "free" in "new" segment list
            for (size_t i = 0; i < m_segments.size(); i++) {
                // Check proportion
                if (m_segmentProportions[i] > 0) {
                    // Necessary segment -> copy to next free slot (only if not already there)
                    if (i > nextFreeSlot) {
                        m_segments[nextFreeSlot] = m_segments[i];
                        m_segmentProportions[nextFreeSlot] = m_segmentProportions[i];
                    }
                    nextFreeSlot++;
                } else {
                    // Unnecessary segment -> do not copy!
                    removedSegments++;
                }
            }

            // Remove garbage data at end of vectors
            if (removedSegments > 0) {
                m_segments.resize(m_segments.size() - removedSegments);
                m_segmentProportions.resize(m_segmentProportions.size() - removedSegments);
            }

            // Pass back count of removed elements
            return removedSegments;
        }

        //! Returns the index of the segment related to the given spline position
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \param [in] splinePosition Position in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \param [out] segmentPosition Position within segment corresponding to the given spline position (optional additional output)
         * \return Index of the segment or -1 in case of an error
         */
        inline int getSegmentIndex(const double& splinePosition, double* const segmentPosition = nullptr) const
        {
            // Abort, if segment count and proportion count do not match (to avoid invalid read access)
            if (m_segments.size() != m_segmentProportions.size())
                return -1;

            // Project position to boundaries
            double projectedSplinePosition = splinePosition;
            if (projectedSplinePosition < 0)
                projectedSplinePosition = 0;
            if (projectedSplinePosition > 1)
                projectedSplinePosition = 1;

            // Search for segment index and segment position (optional)
            double startPosition = 0; // Start position of current segment in spline-coordinates
            double endPosition = 0; // End position of current segment in spline-coordinates
            for (size_t i = 0; i < m_segments.size(); i++) {
                // Compute end position
                endPosition = startPosition + m_segmentProportions[i];
                if (i == m_segments.size() - 1)
                    endPosition = 1.0; // Hard-setting to 1.0 in case of the last segment (to avoid numeric errors)

                // Check, if target position is between start and end
                if (projectedSplinePosition >= startPosition && projectedSplinePosition <= endPosition) {
                    // We found the corresponding segment!

                    // Optional: get segment position from coordinate transformation
                    if (segmentPosition != nullptr) {
                        // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
                        // Linear relationship: segment-position = alpha * spline-position + beta
                        // Derivation: d(segment-position) / d(spline-position) = alpha
                        double alpha = 0;
                        if (endPosition != startPosition)
                            alpha = 1.0 / (endPosition - startPosition);
                        double beta = -alpha * startPosition;

                        // Compute segment position from spline position
                        *segmentPosition = alpha * projectedSplinePosition + beta;
                    }

                    // Stop search and pass back result
                    return (int)i;
                } else {
                    // Otherwise: go to next segment
                    startPosition = endPosition; // Start position of next segment is end position of current segment
                }
            }

            // We could not find the corresponding segment -> error
            return -1;
        }

        //! Evaluation of (derivative of) underlying curve
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \warning For \p derivationOrder > 0 the evaluate function does **not** return the angular velocity/acceleration/...
         * but rather the "pure" derivative of the quaternion with respect to the *spline-coordinates* (**not** relative
         * to *segment-coordinates*)!
         *
         * \param [in] splinePosition Position to evaluate (derivative) at in *spline-coordinates* (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \param [in] derivationOrder Order of the derivation (use 0 for base function, i.e. no derivation)
         * \return Value or derivative at specified position - gradient is expressed in *spline-coordinates*
         */
        virtual Eigen::Quaterniond evaluate(const double& splinePosition, const unsigned int& derivationOrder = 0) const
        {
            // Initialize return value
            Eigen::Quaterniond returnValue(1, 0, 0, 0);

            // Get segment index and position within this segment
            double segmentPosition = 0;
            int segmentIndex = getSegmentIndex(splinePosition, &segmentPosition);

            // Abort, if we did not find a corresponding segment
            if (segmentIndex == -1)
                return returnValue;

            // Check if we only want the base function (for speed up of computation)
            if (derivationOrder == 0) {
                // ... -> just evaluate underlying function of segment and return this value
                return m_segments[segmentIndex].evaluate(segmentPosition, 0);
            } else {
                // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
                // Linear relationship: segment-position = alpha * spline-position + beta
                // Derivation: d(segment-position) / d(spline-position) = alpha
                double alpha = 0;
                if (m_segmentProportions[segmentIndex] > 0)
                    alpha = 1.0 / m_segmentProportions[segmentIndex];

                // Apply coordinate transformation (chain rule) for derivatives
                double chainRuleMultiplier = alpha;
                for (unsigned int i = 2; i <= derivationOrder; i++)
                    chainRuleMultiplier *= alpha;

                // Compute gradient relative to segment coordinates
                returnValue = m_segments[segmentIndex].evaluate(segmentPosition, derivationOrder);

                // Pass transformed gradient
                returnValue.coeffs() *= chainRuleMultiplier;
                return returnValue;
            }
        }

        //! Evaluation of value (D0) and first derivative (D1) of underlying curve (improved performance over calling \ref evaluate() two times)
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \warning Derivatives do **not** represent the angular velocity/acceleration/...
         * but rather the "pure" derivative of the quaternion with respect to the *spline-coordinates* (**not** relative to *segment-coordinates*)!
         *
         * \param [in] splinePosition Position to evaluate (derivative) at in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \param [out] value Value at specified position
         * \param [out] firstDerivative First derivative at specified position - gradient is expressed in *spline-coordinates*
         * \return `true` on success, `false` otherwise
         */
        inline bool evaluateD0D1(const double& splinePosition, Eigen::Quaterniond& value, Eigen::Quaterniond& firstDerivative) const
        {
            // Get segment index and position within this segment
            double segmentPosition = 0;
            int segmentIndex = getSegmentIndex(splinePosition, &segmentPosition);

            // Abort, if we did not find a corresponding segment
            if (segmentIndex == -1)
                return false;

            // Compute value
            value = m_segments[segmentIndex].evaluate(segmentPosition, 0);

            // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
            // Linear relationship: segment-position = alpha * spline-position + beta
            // Derivation: d(segment-position) / d(spline-position) = alpha
            double alpha = 0;
            if (m_segmentProportions[segmentIndex] > 0)
                alpha = 1.0 / m_segmentProportions[segmentIndex];

            // Compute first derviative
            firstDerivative = m_segments[segmentIndex].evaluate(segmentPosition, 1);
            firstDerivative.coeffs() *= alpha;

            // Success
            return true;
        }

        //! Evaluation of value (D0), first (D1) and second derivative (D2) of underlying curve (improved performance over calling \ref evaluate() three times)
        /*!
         * \warning You may call \ref isValid() before to check, if the spline is valid! If the spline is invalid the behaviour of this method is not defined!
         *
         * \warning Derivatives do **not** represent the angular velocity/acceleration/...
         * but rather the "pure" derivative of the quaternion with respect to the *spline-coordinates* (**not** relative to *segment-coordinates*)!
         *
         * \param [in] splinePosition Position to evaluate (derivatives) at in `spline-coordinates` (must be within \f$ [0,\,1] \f$, values are projected otherwise)
         * \param [out] value Value at specified position
         * \param [out] firstDerivative First derivative at specified position - gradient is expressed in *spline-coordinates*
         * \param [out] secondDerivative Second derivative at specified position - gradient is expressed in *spline-coordinates*
         * \return `true` on success, `false` otherwise
         */
        inline bool evaluateD0D1D2(const double& splinePosition, Eigen::Quaterniond& value, Eigen::Quaterniond& firstDerivative, Eigen::Quaterniond& secondDerivative) const
        {
            // Get segment index and position within this segment
            double segmentPosition = 0;
            int segmentIndex = getSegmentIndex(splinePosition, &segmentPosition);

            // Abort, if we did not find a corresponding segment
            if (segmentIndex == -1)
                return false;

            // Compute value
            value = m_segments[segmentIndex].evaluate(segmentPosition, 0);

            // Compute coordinate transformation (spline-coordinates <-> segment-coordinates)
            // Linear relationship: segment-position = alpha * spline-position + beta
            // Derivation: d(segment-position) / d(spline-position) = alpha
            double alpha = 0;
            if (m_segmentProportions[segmentIndex] > 0)
                alpha = 1.0 / m_segmentProportions[segmentIndex];

            // Compute first derviative
            firstDerivative = m_segments[segmentIndex].evaluate(segmentPosition, 1);
            firstDerivative.coeffs() *= alpha;

            // Compute second derivative
            secondDerivative = m_segments[segmentIndex].evaluate(segmentPosition, 2);
            secondDerivative.coeffs() *= (alpha * alpha);

            // Success
            return true;
        }

        //! Interpolate parameters of underlying curve to fit specific constraints (start-end-quaternion, control-points, ...)
        /*!
         * A detailed description on how to set the \p parameters is given below at the corresponding section.
         * \param [in] parameters Interpolation parameters (meaning varies depending on chosen method - for details see corresponding sections below)
         * \param [in] proportions Pointer to segment proportions. If `nullptr`, all segments have the same proportion (uniform distribution), otherwise uses given proportions.
         * \param [in] method Interpolation method to be used (has to be compatible with specified curve type)
         * \param [out] result Pointer to flag in which the result of the algorithm should be stored
         * \return `true` on success, `false` otherwise
         */
        bool interpolate(const std::vector<Eigen::Quaterniond, Eigen::aligned_allocator<Eigen::Quaterniond>>& parameters, const std::vector<double>* const proportions, const QuaternionSplineInterpolationMethod& method, QuaternionSplineResult* const result = nullptr)
        {
            // Assume success
            if (result != nullptr)
                *result = QuaternionSplineResult::SUCCESS;

            // Check, if proportions are specified
            bool proportionsSpecified = false;
            if (proportions != nullptr)
                if (proportions->size() > 0)
                    proportionsSpecified = true;

            // Clear current data
            m_segments.clear();
            if (proportionsSpecified == false)
                m_segmentProportions.clear();
            else
                m_segmentProportions = *proportions;

            /*! Notation:
             *  ---------
             * Symbol        | Code   | Description
             * ------        | -----  | -----------
             * \f$ n \f$     | `n`    | segment count
             * \f$ s_i \f$   | `si`   | spline segment \f$ i \f$
             * \f$ x_i \f$   | `xi`   | *segment-coordinate* of segment \f$ i \f$
             * \f$ h_i \f$   | `hi`   | segment "duration" of segment \f$ i \f$ (= segment proportion)
             * \f$ g_i \f$   | `gi`   | reciprocal of \f$ h_i \f$ (i.e. \f$ g_i = \frac{1}{h_i} \f$)
             * \f$ z \f$     | `z`    | *spline-coordinate* (\f$ z \in [0,\,1] \f$, \f$ z \f$ is **not** the time in general!)
             * \f$ y_i \f$   | `yi`   | value of spline segment \f$ i \f$
             *
             * \par Segmentation
             * \code
             * | ---- s0 ---- | ---- s1 ---- | ---- ... ---- | ---- si ---- | ---- ... ---- | ---- sn-1 ---- |
             * 0 ---- x0 --- 1|0 --- x1 --- 1|0 --- ... --- 1|0 --- xi --- 1|0 --- ... --- 1|0 --- xn-1 --- 1|
             * | ---- h0 ---- | ---- h1 ---- | ---- ... ---- | ---- hi ---- | ---- ... ---- | ---- hn-1 ---- |
             * 0 ------------------------------------------- z --------------------------------------------- 1
             * z0 ----------- z1 ----------- z2 --- ... ---- zi ---------- zi+1 -- ... --- zn-1 ------------ zn
             * y0 ----------- y1 ----------- y2 --- ... ---- yi ---------- yi+1 -- ... --- yn-1 ------------ yn
             * \endcode
             *
             * **Note:** \f$ x_{i+1} \f$ will be written as `xip1` ("p" for "plus") and \f$ x_{i-1} \f$ will be written as `xim1` ("m" for "minus")
             */

            /*! Notes on Coordinate Systems
             *  ---------------------------
             * \par Coordinate transformation
             * The transformation from *spline-coordinates* \f$ z \f$ to *segment-coordinates* \f$x_i \f$ is given by \f$ x_i(z) = \alpha_i\,z + \beta_i \f$.
             *
             * With the boundary values:
             * - Start: \f$ x_i(z_i) = \alpha_i \, z_i + \beta_i = 0 \f$
             * - End: \f$ x_i(z_{i+1}) = \alpha_i\,z_{i+1} + \beta_i = 1 \f$
             *
             * we obtain
             * \f[ \alpha_i = \frac{1}{z_{i+1} - z_i} = \frac{1}{h_i} = g_i \f]
             * and
             * \f[ \beta_i = - \frac{z_i}{h_i} \f]
             */

            // Check interpolation type of underlying curve
            switch (CurveType::interpolationType()) {
            // =================================
            // --- Case: SLERP interpolation ---
            // =================================
            case QuaternionCurve::InterpolationType::SLERP: {
                // Check interpolation method
                switch (method) {
                case QuaternionSplineInterpolationMethod::SLERP: {
                    /*! Interpolation with SLERP
                     *  ------------------------
                     * <b>S</b>pherical <b>L</b>inear int<b>ERP</b>olation (SLERP) between start- and end-quaternions of each segment.
                     * \f$ C^0 \f$ continuous, thus the "velocity" jumps at keyframes!
                     *
                     * (*according to Shoemake, "Animating Rotation with Quaternion Curves", SIGGRAPH Comput. Graph., 1985, volume 19, number 3, ACM (New York, NY, USA), DOI:[10.1145/325165.325242](https://www.doi.org/10.1145/325165.325242), p.245-254*)
                     *
                     * \par Parameter-layout
                     * Element         | Description
                     * -------         | -----------
                     * parameters[0]   | keyframe quaternion at beginning of first segment
                     * ...             | ...
                     * parameters[n-1] | keyframe quaternion at beginning of last segment
                     * parameters[n]   | keyframe quaternion at end of last segment
                     */

                    // Checking parameter set
                    if (parameters.size() <= 1 || (proportionsSpecified == true && m_segmentProportions.size() != parameters.size() - 1)) {
                        if (result != nullptr)
                            *result = QuaternionSplineResult::ERROR_INVALID_PARAMETERS;
                        return false;
                    }

                    // Initialize helpers
                    size_t segmentCount = parameters.size() - 1;

                    // Allocate memory
                    m_segments.resize(segmentCount);
                    if (proportionsSpecified == false) // use uniform distribution
                        m_segmentProportions.resize(segmentCount, (double)1.0 / segmentCount);

                    // Create segments and set *control-point* quaternions
                    for (size_t i = 0; i < segmentCount; i++) {
                        m_segments[i].m_controlPoints[0] = parameters[i];
                        m_segments[i].m_controlPoints[1] = parameters[i + 1];
                    }
                    break;
                }
                default: {
                    // No interpolation methods implemented for now
                    if (result != nullptr)
                        *result = QuaternionSplineResult::ERROR_NOT_IMPLEMENTED;
                    return false;
                }
                }
                break;
            }
            // ==================================
            // --- Case: BEZIER interpolation ---
            // ==================================
            case QuaternionCurve::InterpolationType::BEZIER_CUBIC: {
                // Check interpolation method
                switch (method) {
                case QuaternionSplineInterpolationMethod::BEZIER_CUBIC_CUSTOM: {
                    /*! Interpolation with BEZIER_CUBIC_CUSTOM
                     *  --------------------------------------
                     * Keyframes are connected via quaternion bezier curves. The *control-points* of the bezier curves are user-defined.
                     * The resulting spline is at least \f$ C^0 \f$ continuous.
                     *
                     * (*according to Shoemake, "Animating Rotation with Quaternion Curves", SIGGRAPH Comput. Graph., 1985, volume 19, number 3, ACM (New York, NY, USA), DOI:[10.1145/325165.325242](https://www.doi.org/10.1145/325165.325242), p.245-254*)
                     *
                     * \par Parameter-layout
                     * Element          | Description
                     * -------          | -----------
                     * parameters[0]    | keyframe quaternion at beginning of first segment
                     * ...              | ...
                     * parameters[n-1]  | keyframe quaternion at beginning of last segment
                     * parameters[n]    | keyframe quaternion at end of last segment
                     * parameters[n+1]  | first *control-point* quaternion of first segment
                     * parameters[n+2]  | second *control-point* quaternion of first segment
                     * ...              | ...
                     * parameters[3n-1] | first *control-point* quaternion of last segment
                     * parameters[3n]   | second *control-point* quaternion of last segment
                     */

                    // No interpolation methods implemented for now
                    if (result != nullptr)
                        *result = QuaternionSplineResult::ERROR_NOT_IMPLEMENTED;
                    return false;

                    break;
                }
                case QuaternionSplineInterpolationMethod::BEZIER_CUBIC_SMOOTH: {
                    /*! Interpolation with BEZIER_CUBIC_SMOOTH
                     *  --------------------------------------
                     * Keyframes are connected via quaternion bezier curves. The *control-points* of the bezier curves are chosen automatically.
                     * The resulting spline is at least \f$ C^0 \f$ continuous.
                     *
                     * (*according to Shoemake, "Animating Rotation with Quaternion Curves", SIGGRAPH Comput. Graph., 1985, volume 19, number 3, ACM (New York, NY, USA), DOI:[10.1145/325165.325242](https://www.doi.org/10.1145/325165.325242), p.245-254*)
                     *
                     * \par Parameter-layout
                     * Element          | Description
                     * -------          | -----------
                     * parameters[0]    | keyframe quaternion at beginning of first segment
                     * ...              | ...
                     * parameters[n-1]  | keyframe quaternion at beginning of last segment
                     * parameters[n]    | keyframe quaternion at end of last segment
                     */

                    // No interpolation methods implemented for now
                    if (result != nullptr)
                        *result = QuaternionSplineResult::ERROR_NOT_IMPLEMENTED;
                    return false;

                    break;
                }
                default: {
                    // No interpolation methods implemented for now
                    if (result != nullptr)
                        *result = QuaternionSplineResult::ERROR_NOT_IMPLEMENTED;
                    return false;
                }
                }
                break;
            }
            default: {
                // Unknown function type
                if (result != nullptr)
                    *result = QuaternionSplineResult::ERROR_NOT_IMPLEMENTED;
                return false;
            }
            }

            // Success
            return true;
        }

        // Encoding
        // --------
        //! Encode member data as XML element and add it to the specified stream
        /*!
         * \param [in,out] stream The stream to which the member data should be appended to.
         * \param [in] XMLIndentationLevel Level of indentation in XML format (used to add spaces)
         * \param [in] XMLTabWhiteSpaces Count of whitespaces corresponding to a tab.
         * \param [in] numericFormat C-format specifier for numeric values.
         * \return Count of appended characters (elements in stream).
         */
        virtual io::encoding::CharacterStreamSize encodeToXML(io::encoding::CharacterStream& stream, const uint8_t& XMLIndentationLevel, const size_t& XMLTabWhiteSpaces, const std::string& numericFormat = "%.7g") const
        {
            io::encoding::CharacterStreamSize addedElements = 0;

            // Start XML element
            for (size_t i = 0; i < XMLIndentationLevel * XMLTabWhiteSpaces; i++)
                addedElements += io::encoding::encode(stream, " ");
            addedElements += io::encoding::encode(stream, "<QuaternionSpline");

            // Write attributes
            addedElements += io::encoding::encode(stream, " SegmentProportions=\"");
            for (size_t i = 0; i < m_segmentProportions.size(); i++) {
                if (i > 0)
                    addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_segmentProportions[i], numericFormat);
            }
            addedElements += io::encoding::encode(stream, "\">\n");

            // Write segments
            for (size_t i = 0; i < m_segments.size(); i++)
                addedElements += m_segments[i].encodeToXML(stream, XMLIndentationLevel + 1, XMLTabWhiteSpaces, numericFormat);

            // End XML element
            for (size_t i = 0; i < XMLIndentationLevel * XMLTabWhiteSpaces; i++)
                addedElements += io::encoding::encode(stream, " ");
            addedElements += io::encoding::encode(stream, "</QuaternionSpline>\n");

            // Pass back added elements in stream
            return addedElements;
        }

    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW // <-- Proper 128 bit alignment of member data necessary for Eigen vectorization
    };

} // namespace curve
} // namespace broccoli

#endif // HAVE_EIGEN3
