/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "Curve.hpp"
#include <array>
#include <math.h>

namespace broccoli {
namespace curve {
    //! Class for trigonometric functions. \f$ f(x) = c_0 + c_1\,func(c_2\,x + c_3) \f$
    /*!
     * Type of trigonometric function \f$ func \f$ may be changed at any time (sine, cosine, ...).
     */
    class Trigonometric : public Curve {
    public:
        enum class Type : uint8_t {
            SINE = 0, //!< \f$ \sin(x) \f$
            COSINE, //!< \f$ \cos(x) \f$
            TANGENT, //!< \f$ \tan(x) \f$
            TYPE_COUNT //!< Total count of elements
        };

        //! Returns the string representation of the given function type
        static std::string typeString(const Type& type)
        {
            // Check type
            switch (type) {
            case Type::SINE:
                return "SINE";
            case Type::COSINE:
                return "COSINE";
            case Type::TANGENT:
                return "TANGENT";
            default:
                return "UNKNOWN";
            }
        }

        //! Default constructor
        /*! Initializes to \f$ f(x) = 0 + 1\cdot \sin(1\cdot x + 0) = \sin(x) \f$ */
        Trigonometric()
            : m_type(Type::SINE)
            , m_coefficients({ { 0, 1, 1, 0 } })
        {
        }

        //! Destructor
        virtual ~Trigonometric()
        {
        }

        //! Comparison operator: **equality**
        virtual bool operator==(const Trigonometric& reference) const
        {
            // Compare base class
            if (Curve::operator!=(reference))
                return false;

            // Compare type
            if (m_type != reference.m_type)
                return false;

            // Compare coefficients
            if (m_coefficients != reference.m_coefficients)
                return false;

            // Otherwise -> equality
            return true;
        }

        //! Comparison operator: **inequality**
        virtual bool operator!=(const Trigonometric& reference) const { return !(*this == reference); }

        //! Get type of underlying function
        static FunctionType functionType() { return FunctionType::TRIGONOMETRIC; }

        // Member data
        Type m_type; //!< Type of trigonometric function
        std::array<double, 4> m_coefficients; //!< Coefficients of trigonometric function (\f$ c_i\f$ corresponds to `m_coefficients[i]`)

        // Evaluation of (derivative of) underlying function (see base class for details)
        /*!
         * \copydoc Curve::evaluate()
         * \warning The \f$n\f$-th derivative of tangent (with \f$ n>1 \f$) is not implemented! In this case 0 will be returned.
         */
        virtual double evaluate(const double& position, const unsigned int& derivationOrder = 0) const
        {
            // Check type
            switch (m_type) {
            case Type::SINE:
                // Check if we only want the base function (for speed up of computation)
                if (derivationOrder == 0)
                    return m_coefficients[0] + m_coefficients[1] * sin(m_coefficients[2] * position + m_coefficients[3]);
                else {
                    // Accumulate powers
                    double c2Powered = m_coefficients[2];
                    for (unsigned int i = 1; i < derivationOrder; i++)
                        c2Powered *= m_coefficients[2];

                    // Determine sign: sign = (-1)^(floor(order/2))
                    double sign = 1;
                    if ((((int)(derivationOrder / 2.0)) % 2) == 1)
                        sign = -1;

                    // Determine if we have an even or uneven order of derivation (for selecting sin/cos)
                    if ((derivationOrder % 2) == 0)
                        return sign * m_coefficients[1] * c2Powered * sin(m_coefficients[2] * position + m_coefficients[3]);
                    else
                        return sign * m_coefficients[1] * c2Powered * cos(m_coefficients[2] * position + m_coefficients[3]);
                }
                break;
            case Type::COSINE:
                // Check if we only want the base function (for speed up of computation)
                if (derivationOrder == 0)
                    return m_coefficients[0] + m_coefficients[1] * cos(m_coefficients[2] * position + m_coefficients[3]);
                else {
                    // Accumulate powers
                    double c2Powered = m_coefficients[2];
                    for (unsigned int i = 1; i < derivationOrder; i++)
                        c2Powered *= m_coefficients[2];

                    // Determine sign: sign = (-1)^(floor(order/2+1/2))
                    double sign = 1;
                    if ((((int)(derivationOrder / 2.0 + 0.5)) % 2) == 1)
                        sign = -1;

                    // Determine if we have an even or uneven order of derivation (for selecting sin/cos)
                    if ((derivationOrder % 2) == 0)
                        return sign * m_coefficients[1] * c2Powered * cos(m_coefficients[2] * position + m_coefficients[3]);
                    else
                        return sign * m_coefficients[1] * c2Powered * sin(m_coefficients[2] * position + m_coefficients[3]);
                }
                break;
            case Type::TANGENT:
                // Check if we only want the base function (for speed up of computation)
                if (derivationOrder == 0)
                    return m_coefficients[0] + m_coefficients[1] * tan(m_coefficients[2] * position + m_coefficients[3]);
                else if (derivationOrder == 1) {
                    // First derivative: d/dx tan(x) = 1 + tan(x)^2
                    double tangent = tan(m_coefficients[2] * position + m_coefficients[3]);
                    return m_coefficients[1] * m_coefficients[2] * (1.0 + tangent * tangent);
                } else {
                    // WARNING: n-th derivative of tangent not implemented! Using zero instead!
                    return 0;
                }
                break;
            default:
                // Unknown type...
                return 0;
            }
        }

        // Encoding
        // --------
        // Encode member data as XML element and add it to the specified stream (see base class for details)
        virtual io::encoding::CharacterStreamSize encodeToXML(io::encoding::CharacterStream& stream, const size_t& XMLIndentationLevel, const size_t& XMLTabWhiteSpaces, const std::string& numericFormat = "%.7g") const
        {
            io::encoding::CharacterStreamSize addedElements = 0;

            // Start XML element
            for (size_t i = 0; i < XMLIndentationLevel * XMLTabWhiteSpaces; i++)
                addedElements += io::encoding::encode(stream, " ");
            addedElements += io::encoding::encode(stream, "<Trigonometric");

            // Write type
            addedElements += io::encoding::encode(stream, " Type=\"");
            addedElements += io::encoding::encode(stream, typeString(m_type));

            // Write coefficient data
            addedElements += io::encoding::encode(stream, "\" Coefficients=\"");
            for (size_t i = 0; i < m_coefficients.size(); i++) {
                if (i > 0)
                    addedElements += io::encoding::encode(stream, ",");
                addedElements += io::encoding::encode(stream, (double)m_coefficients[i], numericFormat);
            }
            addedElements += io::encoding::encode(stream, "\">");

            // End XML element
            addedElements += io::encoding::encode(stream, "</Trigonometric>\n");

            // Pass back added elements in stream
            return addedElements;
        }
    };

} // namespace curve
} // namespace broccoli
