/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "LogFileBase.hpp"
#include "LogFileData.hpp"

namespace broccoli {
namespace io {
    //! Template class for logfiles (arbitrary types of logdata)
    /*!
     * The class derives from \ref LogFileBase and adds the missing functionality to process \ref LogFileData objects. The class is
     * implemented as template to allow arbitrary variants of \ref LogFileData (template parameter \p LogFileDataType). The parameter
     * \p LogFileDataType has to be derived from the abstract base class \ref LogFileData.
     */
    template <class LogFileDataType>
    class LogFile : public LogFileBase {
    public:
        // Specialized constructor (see base class for details)
        //! \copydoc LogFileBase::LogFileBase()
        /*! \param [in] bufferSize Size of circular buffer for storing \ref LogFileData objects. A buffer overflow is safe, but implies discarding not written \ref LogFileData objects. */
        LogFile(const std::string& identifier, const std::string& targetDirectory, const std::string& fileName, const std::string& fileExtension, const bool& addTimeStampToName, const bool& compressed, const bool& enableSkipping, const size_t& bufferSize, const size_t& maximumStreamSize = 4096 /* <-- 4kB per default */)
            : LogFileBase(identifier, targetDirectory, fileName, fileExtension, addTimeStampToName, compressed, enableSkipping, maximumStreamSize)
            , m_dataBuffer(bufferSize)
        {
        }

        //! Default constructor (minimal set of parameters)
        /*!
         * \param [in] identifier Initializes: \copybrief identifier()
         * \param [in] bufferSize Size of circular buffer for storing \ref LogFileData objects. A buffer overflow is safe, but implies discarding not written \ref LogFileData objects.
         * \param [in] maximumStreamSize Estimated maximum count of bytes to write at once. Used to pre-allocate memory for better runtime performance.
         */
        LogFile(const std::string& identifier, const size_t& bufferSize, const size_t& maximumStreamSize = 4096 /* <-- 4kB per default */)
            : LogFile(identifier, "", "unknown", ".log", false, false, true, bufferSize, maximumStreamSize)
        {
        }

        // Destructor (see base class for details)
        //! \copydoc ~LogFileBase()
        virtual ~LogFile()
        {
            // Check, if file is open
            if (isOpen() == true) {
                // Try to write footer
                writeLogFooter(core::Time::currentTime());

                // Try to close file
                closeLogFile();
            }

            // Clear buffers
            m_dataBuffer.clear();
        }

        //! **Thread-safe** adding of the given logdata object to the logdata buffer
        /*!
         * Efficient copy to buffer (no new allocations!). May skip, depending on state of \ref active() and \ref countOfObjectsToSkip()
         *
         * \attention The original data is modified (updated timing information).
         *
         * \param [in] newData Logfile data object to add to the buffer.
         * \return `true`, in case the object was added to the buffer, `false` otherwise
         */
        bool addNewDataToBuffer(const LogFileDataType& newData)
        {
            // Update count of received objects
            uint64_t currentTotalReceivedObjectCount = totalReceivedObjectCount() + 1;
            setTotalReceivedObjectCount(currentTotalReceivedObjectCount);

            // Check, if logfile is deactivated
            if (active() == false)
                return false; // ...inactive -> do not add to buffer

            // Check, if we should skip this object
            if (enableSkipping() == false || currentTotalReceivedObjectCount % (countOfObjectsToSkip() + 1) == 0) {
                // ...no skipping -> add object to buffer

                // Set time in new data object to current time (remember this event)
                newData.m_timeAddedToBuffer = core::Time::currentTime();

                // Add new data element to the buffer
                m_dataBuffer.push(newData);

                // Object was added to the buffer
                return true;
            } else
                return false; // Object was skipped
        }

        // Forced clearing of log buffer (thread-safe) (see base class for details)
        virtual void clearLogBuffer()
        {
            // Clear all elements
            m_dataBuffer.clear();

            // Reset counter for received objects
            setTotalReceivedObjectCount(0);
        }

        // Thread-safe writing of the file header (see base class for details)
        virtual bool writeLogHeader(const core::Time& logStartTime)
        {
            // Lock file-mutex for write
            pthread_rwlock_wrlock(&m_mutexFile);

            // Check, if file is open
            // ----------------------
            if (m_isOpen == false) {
                m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_WRITE_FAILED_NOTOPEN));
                pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                return false;
            }

            // Write header
            // ------------
            LogFileDataType dummyDataForHeader; // Data dummy object only used to create the corresponding header of the logfile
            dummyDataForHeader.encodeHeaderToLogStream(m_logStream, logStartTime, identifier());

            // Write stream to file
            if (writeCurrentLogStreamToFile() == false) {
                // Writing header failed -> try to close file
                pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                closeLogFile();
                return false;
            }

            // Unlock file-mutex
            pthread_rwlock_unlock(&m_mutexFile);

            // Success!
            return true;
        }

        // Write current contents of buffer to file (thread-safe) (see base class for details)
        virtual bool processLogBuffer()
        {
            // Lock file-mutex for write
            pthread_rwlock_wrlock(&m_mutexFile);

            // Check, if file is open
            // (If file is not open we can not write to it -> skip but keep data in buffer (in case we want to dump the data later))
            if (m_isOpen == true) {
                // Encode elements in buffer and add them to the stream
                LogFileDataType currentElement; // Allocated on stack -> fast!
                while (m_dataBuffer.pop(currentElement) > 0) // Try to get next element (stops if there are no elements left)
                {
                    // Update counter of written elements
                    setTotalWrittenObjectCount(totalWrittenObjectCount() + 1);

                    // Encode elements
                    currentElement.encodeDataToLogStream(m_logStream);
                }

                // Write stream to file
                if (writeCurrentLogStreamToFile() == false) {
                    // Writing header failed -> try to close file
                    pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                    closeLogFile();
                    return false;
                }
            }

            // Unlock file-mutex
            pthread_rwlock_unlock(&m_mutexFile);

            // Success
            return true;
        }

        // Thread-safe writing of the file footer (see base class for details)
        virtual bool writeLogFooter(const core::Time& logStopTime)
        {
            // Lock file-mutex for write
            pthread_rwlock_wrlock(&m_mutexFile);

            // Check, if file is open
            // ----------------------
            if (m_isOpen == false) {
                m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_WRITE_FAILED_NOTOPEN));
                pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                return false;
            }

            // Write footer
            // ------------
            LogFileDataType dummyDataForFooter; // Data dummy object only used to create the corresponding footer of the logfile
            dummyDataForFooter.encodeFooterToLogStream(m_logStream, logStopTime, identifier());

            // Write stream to file
            if (writeCurrentLogStreamToFile() == false) {
                // Writing footer failed -> try to close file
                pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                closeLogFile();
                return false;
            }

            // Unlock file-mutex
            pthread_rwlock_unlock(&m_mutexFile);

            // Success!
            return true;
        }

        // Count of objects currently in data buffer (see base class for details)
        virtual inline uint64_t countOfObjectsInBuffer() const { return m_dataBuffer.elementCount(); }

    protected:
        memory::CircularBuffer<LogFileDataType> m_dataBuffer; //!< Efficient **thread-safe** circular buffer for logfile data. Implements its own mutex.
    };
} // namespace io
} // namespace broccoli
