/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "../memory/CircularBuffer.hpp"
#include "LogFileData.hpp"
#include "LogFileSignal.hpp"
#include "filesystem.hpp"
#include <fstream>
#include <iostream>
#include <math.h>
#include <pthread.h>
#include <sstream>
#include <stdint.h>
#include <string>
#include <sys/stat.h>
#include <sys/types.h>
#ifdef HAVE_ZLIB // is zlib library available?
#include <zlib.h>
#endif // HAVE_ZLIB

namespace broccoli {
namespace io {
    //! Abstract base class of logfile template
    /*!
     * Necessary to allow management of multiple logfiles (with different \ref LogFileData types)
     * by one \ref Logger. Contains pure virtual functions
     *
     *  - \ref clearLogBuffer()
     *  - \ref writeLogHeader()
     *  - \ref processLogBuffer()
     *  - \ref writeLogFooter()
     *
     * which have to be defined by the derived class since they depend on the chosen type of \ref LogFileData. Those
     * methods are triggered by the \ref Logger and thus have to be known by the \ref Logger at compile time. In contrast
     * it is not neccessary to know the exact type of the logdata at this point.
     *
     * The class further handles all neccessary file and parameter members including corresponding **thread-safe** getters
     * and setters.
     *
     * \remark The class includes a signal buffer (\ref m_signalBuffer) which allows the \ref Logger to pass through signals
     * of the \ref LogFile to a supervising observer (i.e. the user application).
     */
    class LogFileBase {
    public:
        //! Specialized constructor
        /*!
         * \param [in] identifier Initializes: \copybrief identifier()
         * \param [in] targetDirectory Initializes the **current** and **desired** target directory containing the logfile
         * \param [in] fileName Initializes the **current** and **desired** name of the logfile (without extension)
         * \param [in] fileExtension Initializes the **current** and **desired** file extension (including dot)
         * \param [in] addTimeStampToName Initializes the **current** and **desired** setting for adding a timestamp (system time) to the filename (after name, before extension)
         * \param [in] compressed Initializes the **current** and **desired** setting for compression of the file (if `true` the extension ".gz" is added to the filepath)
         * \param [in] enableSkipping Initializes: \copybrief enableSkipping()
         * \param [in] maximumStreamSize Estimated maximum count of bytes to write at once. Used to pre-allocate memory for better runtime performance.
         */
        LogFileBase(const std::string& identifier, const std::string& targetDirectory, const std::string& fileName, const std::string& fileExtension, const bool& addTimeStampToName, const bool& compressed, const bool& enableSkipping, const size_t& maximumStreamSize = 4096 /* <-- 4kB per default */)
            : m_identifier(identifier)
            , m_mutex(PTHREAD_RWLOCK_INITIALIZER)
            , m_mutexFile(PTHREAD_RWLOCK_INITIALIZER)
            , m_desiredTargetDirectory(targetDirectory)
            , m_desiredFileName(fileName)
            , m_desiredFileExtension(fileExtension)
            , m_desiredAddTimeStampToName(addTimeStampToName)
            , m_desiredCompressed(compressed)
            , m_active(true)
            , m_enableSkipping(enableSkipping)
            , m_countOfObjectsToSkip(0)
            , m_targetDirectory(targetDirectory)
            , m_fileName(fileName)
            , m_fileExtension(fileExtension)
            , m_addTimeStampToName(addTimeStampToName)
            , m_compressed(compressed)
            , m_totalReceivedObjectCount(0)
            , m_totalWrittenObjectCount(0)
            , m_isOpen(false)
            , m_totalBytesWritten(0)
            , m_lastFlushByteCount(0)
            , m_signalBuffer(1000)
        {
            // Pre-allocate memory (estimate memory requirements)
            m_logStream.reserve(maximumStreamSize);
        }

        //! Deleted copy constructor (prevent user to copy object - would be dangerous due to internal mutexes)
        LogFileBase(const LogFileBase& original) = delete;

        //! Deleted copy assignment operator (prevent user to copy object - would be dangerous due to internal mutexes)
        LogFileBase& operator=(const LogFileBase& reference) = delete;

        //! Destructor
        /*! Closes the file and clears buffers. */
        virtual ~LogFileBase()
        {
            // Check, if file is open
            if (isOpen() == true) {
                // Try to close file
                closeLogFile();
            }

            // Clear buffers
            m_signalBuffer.clear();
        }

        // Main functionality (triggered by Logger)
        // ------------------
        //! **Thread-safe** opening of new log file to write into (triggered by supervising \ref Logger)
        /*!
         * \remark In case of an error a corrresponding \ref LogFileSignal is added to the buffer (\ref m_signalBuffer).
         *
         * \return `true` on success, `false` otherwise
         */
        virtual bool openLogFile()
        {
            // Lock file-mutex for write
            pthread_rwlock_wrlock(&m_mutexFile);

            // Check, if file has not been opened already
            // ------------------------------------------
            if (m_isOpen == true) {
                m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_OPEN_FAILED_ALREADY_OPEN));
                pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                return false;
            }

            // Update file information
            // -----------------------
            // Update all within one mutex-lock to avoid inconsistent data
            pthread_rwlock_wrlock(&m_mutex); // Lock parameter-mutex for write
            m_targetDirectory = m_desiredTargetDirectory;
            m_fileName = m_desiredFileName;
            m_fileExtension = m_desiredFileExtension;
            m_addTimeStampToName = m_desiredAddTimeStampToName;
            m_compressed = m_desiredCompressed;
            std::string currentTargetDirectory = m_targetDirectory;
            std::string currentFileName = m_fileName;
            std::string currentFileExtension = m_fileExtension;
            bool currentAddTimeStampToName = m_addTimeStampToName;
            bool currentCompressed = m_compressed;
            pthread_rwlock_unlock(&m_mutex); // Unlock parameter-mutex

            // Compute new total filepath
            // --------------------------
            std::stringstream filePathStream;

            // Add target directory
            if (currentTargetDirectory.length() > 0)
                filePathStream << currentTargetDirectory << "/";

            // Add file name
            filePathStream << currentFileName;

            // Optionally add time stamp
            if (currentAddTimeStampToName == true)
                filePathStream << "-" << core::Time::currentTime().encodeToDateTimeString(true, "%Y-%m-%d_%H-%M-%S", 0);

            // Add extension
            filePathStream << currentFileExtension;

            // Add extension of compression file format
            if (currentCompressed == true)
                filePathStream << ".gz";

            // Convert to string
            std::string filePath = filePathStream.str();

            // Check, if target directory exists and try to create it if not
            // -------------------------------------------------------------
            // Try to create folder if it does not exist already
            if (filesystem::directoryExists(currentTargetDirectory) == false) {
                (void)filesystem::createDirectory(currentTargetDirectory); // May fail, if another thread/process creates the directory in the meantime -> return value is not representative

                // Check, if directory exists now
                if (filesystem::directoryExists(currentTargetDirectory) == false) {
                    // We could not create the directory so we also cannot open the log file -> abort
                    m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_OPEN_FAILED_FOLDER_CREATION_FAILED));
                    pthread_rwlock_unlock(&m_mutexFile); // Unlock file-mutex
                    return false;
                }
            }

            // Open file
            // ---------
            // Check, if we should use file compression
            if (currentCompressed == false) {
                // Try to open uncompressed file
                try {
                    m_uncompressedFile = fopen(filePath.c_str(), "w");
                    if (m_uncompressedFile != nullptr)
                        m_isOpen = true;
                } catch (...) {
                    m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_OPEN_FAILED_LOWLEVEL));
                    pthread_rwlock_unlock(&m_mutex); // Unlock file-mutex
                    return false;
                }
            } else {
                // Try to open compressed file
#ifdef HAVE_ZLIB
                try {
                    m_compressedFile = gzopen(filePath.c_str(), "w");
                    if (m_compressedFile != nullptr)
                        m_isOpen = true;
                } catch (...) {
                    m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_OPEN_FAILED_LOWLEVEL));
                    pthread_rwlock_unlock(&m_mutex); // Unlock file-mutex
                    return false;
                }
#else
                // Not compiled with zlib support
                m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_ZLIB_NOT_FOUND));
                pthread_rwlock_unlock(&m_mutex); // Unlock file-mutex
                return false;
#endif
            }

            // Reset byte counters
            m_totalBytesWritten = 0;
            m_lastFlushByteCount = 0;

            // Unlock file-mutex
            pthread_rwlock_unlock(&m_mutexFile);

            // Reset counter for written objects
            setTotalWrittenObjectCount(0);

            // Success!
            return true;
        }

        //! **Thread-safe** closing of logfile (triggered by supervising \ref Logger)
        /*! \return `true` if file was open, `false` if it was not open */
        virtual bool closeLogFile()
        {
            // Get compression state
            bool currentCompressed = compressed();

            // Lock file-mutex for write
            pthread_rwlock_wrlock(&m_mutexFile);

            // Remember current state
            bool wasFileOpen = m_isOpen;

            // Close file (only if it was open)
            // ----------
            if (wasFileOpen == true) {
                try {
                    if (currentCompressed == false) {
                        fclose(m_uncompressedFile);
                    }
#ifdef HAVE_ZLIB
                    else {
                        gzclose(m_compressedFile);
                    }
#endif

                } catch (...) {
                    // ...do nothing in case of a lowlevel error
                }
            }

            // Update state
            m_isOpen = false;

            // Unlock file-mutex
            pthread_rwlock_unlock(&m_mutexFile);

            // Return previous state
            return wasFileOpen;
        }

        // Abstract implementations
        // ------------------------
        //! **Thread-safe** forced clearing of logdata buffer
        /*! Clears the complete logdata buffer. \warning All data in the buffer is discarded! */
        virtual void clearLogBuffer() = 0;

        //! **Thread-safe** writing of the file header
        /*!
         * \remark In case of an error a corrresponding \ref LogFileSignal is added to the buffer (\ref m_signalBuffer).
         *
         * \param [in] logStartTime The time when the log was started.
         * \return `true` on success, `false` otherwise
         */
        virtual bool writeLogHeader(const core::Time& logStartTime) = 0;

        //! **Thread-safe** writing of current contents of the logdata buffer to the logfile
        /*!
         * \remark In case of an error a corrresponding \ref LogFileSignal is added to the buffer (\ref m_signalBuffer).
         *
         * \return `true` on success, `false` otherwise
         */
        virtual bool processLogBuffer() = 0;

        //! **Thread-safe** writing of the file footer
        /*!
         * \remark In case of an error a corrresponding \ref LogFileSignal is added to the buffer (\ref m_signalBuffer).
         *
         * \param [in] logStopTime The time when the log was stopped.
         * \return `true` on success, `false` otherwise
         */
        virtual bool writeLogFooter(const core::Time& logStopTime) = 0;

    protected:
        //! Write current content of log stream to file (**not** thread-safe, file-mutex has to be locked by caller!)
        /*!
         * \warning **Not** thread-safe, file-mutex (\ref m_mutexFile) has to be locked by caller!
         *
         * \remark In case of an error a corrresponding \ref LogFileSignal is added to the buffer (\ref m_signalBuffer).
         *
         * \return `true` on success, `false` otherwise
         */
        bool writeCurrentLogStreamToFile()
        {
            // Get compression state
            bool currentCompressed = compressed();

            // Only write, if there is really something to write
            if (m_logStream.size() == 0)
                return true; // Nothing to do -> handle as success

            // Step 1: write data
            // ------------------
            try {
                if (currentCompressed == false)
                    fprintf(m_uncompressedFile, "%.*s", (int)m_logStream.size(), &m_logStream[0]);
#ifdef HAVE_ZLIB
                else {
                    // Stream splitting
                    // ----------------
                    // gzprintf can only take a maximum of 8191 bytes (2^13 - 1 (null-terminated))
                    // -> stream has to be split into parts, if it is greater

                    // Compute in how many parts we have to split
                    uint64_t partCount = ::ceil((double)m_logStream.size() / 8191.0);
                    // Write all parts except for the last one
                    for (uint64_t part = 0; part < partCount - 1; part++)
                        if (gzprintf(m_compressedFile, "%.*s", (int)8191, &m_logStream[part * 8191]) != 8191) {
                            m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_WRITE_FAILED_DATABUFFER));
                            return false;
                        }

                    // Write last part
                    if (gzprintf(m_compressedFile, "%.*s", (int)(m_logStream.size() - (partCount - 1) * 8191), &m_logStream[(partCount - 1) * 8191]) != (int)(m_logStream.size() - (partCount - 1) * 8191)) {
                        m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_WRITE_FAILED_DATABUFFER));
                        return false;
                    }
                }
#endif
            } catch (...) {
                m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_WRITE_FAILED_LOWLEVEL));
                return false;
            }

            // Update counter of totally written bytes (used for flush control)
            m_totalBytesWritten += m_logStream.size();

            // Clear data but keep capacity -> reusing already allocated memory!
            m_logStream.clear();

            // Step 2: flush data
            // ------------------
            // Avoid to call flush too often, since this reduces the compression level in the compressed case
            try {
                if (m_totalBytesWritten >= m_lastFlushByteCount + 4096) // Flush every 4 kB
                {
                    // Flush
                    if (currentCompressed == false)
                        fflush(m_uncompressedFile);
#ifdef HAVE_ZLIB
                    else
                        gzflush(m_compressedFile, Z_SYNC_FLUSH);
#endif

                    // Remember event
                    m_lastFlushByteCount = m_totalBytesWritten;
                }
            } catch (...) {
                m_signalBuffer.push(LogFileSignal(LogFileSignal::Type::ERROR, LogFileSignal::Message::FILE_WRITE_FAILED_FLUSH));
                return false;
            }

            // Success
            return true;
        }

        // Constants
        // ---------
    private:
        const std::string m_identifier; //!< \copybrief identifier()

        // Protected data
        // --------------
    private:
        mutable pthread_rwlock_t m_mutex; //!< Read-write mutex for thread-safety of common parameters
    protected:
        mutable pthread_rwlock_t m_mutexFile; //!< Read-write mutex for thread-safety of file access

    private:
        // Start protected area of m_mutex (external-visible parameters)
        std::string m_desiredTargetDirectory; //!< \copybrief desiredTargetDirectory()
        std::string m_desiredFileName; //!< \copybrief desiredFileName()
        std::string m_desiredFileExtension; //!< \copybrief desiredFileExtension()
        bool m_desiredAddTimeStampToName; //!< \copybrief desiredAddTimeStampToName()
        bool m_desiredCompressed; //!< \copybrief desiredCompressed()
        bool m_active; //!< \copybrief active()
        bool m_enableSkipping; //!< \copybrief enableSkipping()
        uint64_t m_countOfObjectsToSkip; //!< \copybrief countOfObjectsToSkip()
        // End protected area of m_mutex (external-visible parameters)

    private:
        // Start protected area of m_mutex (internal-only parameters)
        std::string m_targetDirectory; //!< \copybrief targetDirectory()
        std::string m_fileName; //!< \copybrief fileName()
        std::string m_fileExtension; //!< \copybrief fileExtension()
        bool m_addTimeStampToName; //!< \copybrief addTimeStampToName()
        bool m_compressed; //!< \copybrief compressed()
        uint64_t m_totalReceivedObjectCount; //!< \copybrief totalReceivedObjectCount()
        uint64_t m_totalWrittenObjectCount; //!< \copybrief totalWrittenObjectCount()
        // End protected area of m_mutex (internal-only parameters)

    protected:
        // Start protected area of m_mutexFile ("internal-only")
        LogFileData::LogStream m_logStream; //!< Pre-allocated data stream (for reusing already allocated memory)
        FILE* m_uncompressedFile; //!< File object for writing (uncompressed)
#ifdef HAVE_ZLIB
        gzFile m_compressedFile; //!< File object for writing (compressed)
#endif
        bool m_isOpen; //!< \copybrief isOpen()
        uint64_t m_totalBytesWritten; //!< Count of total bytes written (uncompressed)
        uint64_t m_lastFlushByteCount; //!< Count of total bytes written at last flush event (used for flush control)
        // End protected area of m_mutexFile ("internal-only")

        // Buffers
        // -------
    public:
        memory::CircularBuffer<LogFileSignal> m_signalBuffer; //!< Buffer for signals of the logfile to a supervising \ref Logger (**thread-safe** on its own)

    private:
        //! Setter template for protected data of common parameter mutex
        /*!
         * Sets the value of a protected data member. Thread-safety is guaranteed by the common parameter mutex.
         * \param [in] newValue New value for data member.
         * \param [in] data Pointer to data member
         */
        template <typename T>
        void setProtectedData(const T& newValue, T* data)
        {
            pthread_rwlock_wrlock(&m_mutex); // Lock protected memory for write
            *data = newValue; // Set value
            pthread_rwlock_unlock(&m_mutex); // Unlock protected memory
        }

        //! Getter template for protected data of common parameter mutex
        /*!
         * Gets the value of a protected data member. Thread-safety is guaranteed by the common parameter mutex.
         * \param [in] data Pointer to data member
         * \return [out] Value of data member (copy)
         */
        template <typename T>
        T getProtectedData(const T* data) const
        {
            pthread_rwlock_rdlock(&m_mutex); // Lock protected memory for read
            T returnValue = *(data); // Store data to temporary variable
            pthread_rwlock_unlock(&m_mutex); // Unlock protected memory
            return returnValue; // Return temporary variable
        }

        // Getters ("external" data)
        // -------
    public:
        //! Unique identifier for this log file \details **Thread-safe getter**
        inline std::string identifier() const { return m_identifier; }
        //! The **desired** target directory containing the logfile (will be used on next \ref openLogFile() call) \details **Thread-safe getter**
        inline std::string desiredTargetDirectory() const { return getProtectedData<std::string>(&m_desiredTargetDirectory); }
        //! The **desired** name of the logfile (without extension) (will be used on next \ref openLogFile() call) \details **Thread-safe getter**
        inline std::string desiredFileName() const { return getProtectedData<std::string>(&m_desiredFileName); }
        //! The **desired** file extension (including dot) (will be used on next \ref openLogFile() call) \details **Thread-safe getter**
        inline std::string desiredFileExtension() const { return getProtectedData<std::string>(&m_desiredFileExtension); }
        //! The **desired** setting for adding a timestamp (system time) to the filename (after name, before extension) (will be used on next \ref openLogFile() call) \details **Thread-safe getter**
        inline bool desiredAddTimeStampToName() const { return getProtectedData<bool>(&m_desiredAddTimeStampToName); }
        //! The **desired** setting for compression of the file (if `true` the extension ".gz" is added to the filepath) (will be used on next \ref openLogFile() call) \details **Thread-safe getter**
        inline bool desiredCompressed() const { return getProtectedData<bool>(&m_desiredCompressed); }
        //! Flag to (de-)activate the logdata buffer. If set to `false`, incoming logdata-objects are not added to the buffer (discards objects immediately before adding to the buffer) \details **Thread-safe getter**
        inline bool active() const { return getProtectedData<bool>(&m_active); }
        //! Flag to (de-)activate skipping of objects (count of objects to skip is determined by \ref countOfObjectsToSkip()) \details **Thread-safe getter**
        inline bool enableSkipping() const { return getProtectedData<bool>(&m_enableSkipping); }
        //! Count of objects to skip (imagine as "write only every i-th dataset") (discards objects immediately before adding to the buffer) \details **Thread-safe getter**
        inline uint64_t countOfObjectsToSkip() const { return getProtectedData<uint64_t>(&m_countOfObjectsToSkip); }
        //! Count of objects currently in data buffer \details **Thread-safe getter**
        virtual inline uint64_t countOfObjectsInBuffer() const = 0;

        // Getters ("internal" data)
        // -------
    public:
        //! The **current** target directory containing the logfile \details **Thread-safe getter**
        inline std::string targetDirectory() const { return getProtectedData<std::string>(&m_targetDirectory); }
        //! The **current** name of the logfile (without extension) \details **Thread-safe getter**
        inline std::string fileName() const { return getProtectedData<std::string>(&m_fileName); }
        //! The **current** file extension (including dot) \details **Thread-safe getter**
        inline std::string fileExtension() const { return getProtectedData<std::string>(&m_fileExtension); }
        //! The **current** setting for adding a timestamp (system time) to the filename (after name, before extension) \details **Thread-safe getter**
        inline bool addTimeStampToName() const { return getProtectedData<bool>(&m_addTimeStampToName); }
        //! The **current** setting for compression of the file (if `true` the extension ".gz" is added to the filepath) \details **Thread-safe getter**
        inline bool compressed() const { return getProtectedData<bool>(&m_compressed); }
        //! Total count of **received** objects (some of them may have been skipped/discarded). Gets resetted with \ref clearLogBuffer() \details **Thread-safe getter**
        inline uint64_t totalReceivedObjectCount() const { return getProtectedData<uint64_t>(&m_totalReceivedObjectCount); }
        //! Total count of **written** objects. Gets resetted with \ref openLogFile() \details **Thread-safe getter**
        inline uint64_t totalWrittenObjectCount() const { return getProtectedData<uint64_t>(&m_totalWrittenObjectCount); }
        //! Flag to indicate if the the file is **currently** open \details **Thread-safe getter**
        inline bool isOpen() const
        {
            pthread_rwlock_rdlock(&m_mutexFile); // Lock protected memory for read
            bool returnValue = m_isOpen; // Store data to temporary variable
            pthread_rwlock_unlock(&m_mutexFile); // Unlock protected memory
            return returnValue; // Return temporary variable
        }

        // Setters ("external" data)
        // -------
    public:
        //! **Thread-safe setter for:** \copybrief desiredTargetDirectory()
        inline void setDesiredTargetDirectory(const std::string& newValue) { setProtectedData<std::string>(newValue, &m_desiredTargetDirectory); }
        //! **Thread-safe setter for:** \copybrief desiredFileName()
        inline void setDesiredFileName(const std::string& newValue) { setProtectedData<std::string>(newValue, &m_desiredFileName); }
        //! **Thread-safe setter for:** \copybrief desiredFileExtension()
        inline void setDesiredFileExtension(const std::string& newValue) { setProtectedData<std::string>(newValue, &m_desiredFileExtension); }
        //! **Thread-safe setter for:** \copybrief desiredAddTimeStampToName()
        inline void setDesiredAddTimeStampToName(const bool& newValue) { setProtectedData<bool>(newValue, &m_desiredAddTimeStampToName); }
        //! **Thread-safe setter for:** \copybrief desiredCompressed()
        inline void setDesiredCompressed(const bool& newValue) { setProtectedData<bool>(newValue, &m_desiredCompressed); }
        //! **Thread-safe setter for:** \copybrief active()
        inline void setActive(const bool& newValue) { setProtectedData<bool>(newValue, &m_active); }
        //! **Thread-safe setter for:**\copybrief enableSkipping()
        inline void setEnableSkipping(const bool& newValue) { setProtectedData<bool>(newValue, &m_enableSkipping); }
        //! **Thread-safe setter for:**\copybrief countOfObjectsToSkip()
        inline void setCountOfObjectsToSkip(const uint64_t& newValue) { setProtectedData<uint64_t>(newValue, &m_countOfObjectsToSkip); }

        // Setters ("internal" data)
        // -------
    protected:
        // "Internal" data
        //! **Thread-safe setter for:** \copybrief targetDirectory()
        inline void setTargetDirectory(const std::string& newValue) { setProtectedData<std::string>(newValue, &m_targetDirectory); }
        //! **Thread-safe setter for:** \copybrief fileName()
        inline void setFileName(const std::string& newValue) { setProtectedData<std::string>(newValue, &m_fileName); }
        //! **Thread-safe setter for:** \copybrief fileExtension()
        inline void setFileExtension(const std::string& newValue) { setProtectedData<std::string>(newValue, &m_fileExtension); }
        //! **Thread-safe setter for:** \copybrief addTimeStampToName()
        inline void setAddTimeStampToName(const bool& newValue) { setProtectedData<bool>(newValue, &m_addTimeStampToName); }
        //! **Thread-safe setter for:** \copybrief compressed()
        inline void setCompressed(const bool& newValue) { setProtectedData<bool>(newValue, &m_compressed); }
        //! **Thread-safe setter for:** \copybrief totalReceivedObjectCount()
        inline void setTotalReceivedObjectCount(const uint64_t& newValue) { setProtectedData<uint64_t>(newValue, &m_totalReceivedObjectCount); }
        //! **Thread-safe setter for:** \copybrief totalWrittenObjectCount()
        inline void setTotalWrittenObjectCount(const uint64_t& newValue) { setProtectedData<uint64_t>(newValue, &m_totalWrittenObjectCount); }
    };
} // namespace io
} // namespace broccoli
