/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#pragma once // Load this file only once

#include "../core/Time.hpp"
#include "encoding.hpp"
#ifdef HAVE_EIGEN3
#include <Eigen/StdVector> // Necessary for 128bit alginnment and STL containers (for vectorization)
#endif // HAVE_EIGEN3
#include <stdio.h>
#include <string>
#include <vector>

namespace broccoli {
namespace io {
    //! Generic log data container (abstract class)
    /*!
     * Pure abstract class. Can be used as base class for log-data containers. Encodes its own data
     * members for writing to a log file (text based, **not** binary). This is done by \ref encodeDataToLogStream()
     * which has to be implemented by the derived class.
     *
     * Furthermore one can overwrite \ref encodeHeaderToLogStream() and \ref encodeFooterToLogStream() to
     * specify how the file header and footer should look like. In the base implementation a default header
     * \code
     * # log started on 20xx-xx-xx xx:xx:xx
     * # log for <logFileName>
     * # <\ref columnHeaderNames()[0]><\ref columnHeaderDelimiter()>0<\ref columnHeaderSeparator()><\ref columnHeaderNames()[1]><\ref columnHeaderDelimiter()>1<\ref columnHeaderSeparator()>...
     * \endcode
     * (third line is skipped if \ref columnHeaderNames() returns an empty list) and footer (empty by default) are predefined.
     *
     * The methods
     * - \ref encodeHeaderToLogStream()
     * - \ref encodeDataToLogStream()
     * - \ref encodeFooterToLogStream()
     *
     * are handled by the corresponding \ref LogFile which in turn is supervised by a \ref Logger. This way the
     * \ref Logger can execute the encoding efficiently in a background thread.
     *
     * \attention
     * It is important to provide a default constructor which does **not** require any parameters. This is necessary
     * since the corresponding \ref Logger will create "dummy" \p LogFileData objects just to be able to encode
     * the file header and footer. Because the \ref Logger does not know the type of the \ref LogFileData at compile
     * time a "trivial" default constructor has to be provided.
     */
    class LogFileData {
    public:
        // Type definitions
        typedef encoding::CharacterStream LogStream; //!< Specifies the used type for storing log stream data
        typedef encoding::CharacterStreamSize LogStreamSize; //!< Specifies the used type for describing the size of a log stream

        //! Default constructor
        /*!
         * \attention Derived classes must also provide a default constructor which does not require any parameters!
         * (see detailed class description of \ref LogFileData for details)
         */
        LogFileData()
            : m_timeAddedToBuffer(-1)
        {
        }

        //! Destructor
        virtual ~LogFileData()
        {
        }

        // Members
        mutable core::Time m_timeAddedToBuffer; //!< Indicates when (system time) the data has been added to the buffer of the corresponding \ref LogFile. For \f$ <0 \f$: data has not been added to the buffer yet

        //! Encode file header to log stream
        /*!
         * The header is written **before** member data is written to the stream. The header is written only **once**.
         * \param [in,out] stream The logging stream to which the header should be appended to.
         * \param [in] logStartTime The time when the log was started.
         * \param [in] logFileIdentifier Identifier of the logfile this data is written to.
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeHeaderToLogStream(LogStream& stream, const core::Time& logStartTime, const std::string& logFileIdentifier = "") const
        {
            LogStreamSize addedElements = 0;

            // Append basic header to stream
            addedElements += encoding::encode(stream, "# log started on ");
            addedElements += encoding::encode(stream, logStartTime.encodeToDateTimeString(true, "%Y-%m-%d %H:%M:%S", 9));
            if (logFileIdentifier.size() > 0) {
                addedElements += encoding::encode(stream, "\n# log for ");
                addedElements += encoding::encode(stream, logFileIdentifier);
            }
            addedElements += encoding::encode(stream, '\n');

            // Append column header
            std::vector<std::string> columnNames = columnHeaderNames();
            if (columnNames.size() > 0) {
                // Start column header
                addedElements += encoding::encode(stream, "# ");

                // Add header for each column
                for (size_t i = 0; i < columnNames.size(); i++) {
                    // Encode name
                    addedElements += encoding::encode(stream, columnNames[i]);

                    // Encode delimiter
                    addedElements += encodeColumnHeaderDelimiter(stream, i);

                    // Encode index
                    addedElements += encoding::encode(stream, (uint64_t)i);

                    // Encode seperator (only if there is a pending column
                    if (i < columnNames.size() - 1)
                        addedElements += encodeColumnHeaderSeparator(stream, i);
                }

                // Finish column header
                addedElements += encoding::encode(stream, '\n');
            }

            // Pass back added elements in stream
            return addedElements;
        }

        //! Encode member data to log stream
        /*!
         * Encodes member data and appends it to the specified stream.
         * \param [in,out] stream The logging stream to which the member data should be appended to.
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeDataToLogStream(LogStream& stream) const = 0;

        //! Encode file footer to log stream
        /*!
         * The footer is written **after** member data is written to the stream. The footer is written only **once**.
         * \param [in,out] stream The logging stream to which the footer should be appended to.
         * \param [in] logStopTime The time when the log was stopped.
         * \param [in] logFileIdentifier Identifier of the logfile this data is written to.
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeFooterToLogStream(LogStream& stream, const core::Time& logStopTime, const std::string& logFileIdentifier = "") const
        {
            // Default: no footer...
            (void)stream; // ...unused
            (void)logStopTime; // ...unused
            (void)logFileIdentifier; // ...unused
            return 0;
        }

        // Encoding of default column headers
        // ----------------------------------
        //! Specification of column header names ("titles" of columns)
        /*! \return List of column header names or empty list in case the column header should be skipped */
        virtual std::vector<std::string> columnHeaderNames() const
        {
            // DUMMY IMPLEMENTATION: empty list to skip output of column header
            return std::vector<std::string>();
        }

        //! Encoding of column header delimiter
        /*!
         * \param [in,out] stream The logging stream to which the column header delimiter should be appended to.
         * \param [in] index Index of the **current** column
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeColumnHeaderDelimiter(LogStream& stream, const size_t& index) const
        {
            // Index unused per default (same delimiter for all columns)
            (void)index;

            // Use default character
            return encoding::encode(stream, (char)':');
        }

        //! Encoding of column header separator
        /*!
         * \param [in,out] stream The logging stream to which the column header separator should be appended to.
         * \param [in] index Index of the **preceding** column
         * \return Count of appended characters (elements added to stream).
         */
        virtual LogStreamSize encodeColumnHeaderSeparator(LogStream& stream, const size_t& index) const
        {
            // Index unused per default (same separator for all columns)
            (void)index;

            // Use default character
            return encoding::encode(stream, (char)' ');
        }

#ifdef HAVE_EIGEN3
    public:
        EIGEN_MAKE_ALIGNED_OPERATOR_NEW // <-- Proper 128 bit alignment of member data necessary for Eigen vectorization
#endif // HAVE_EIGEN3
    };
} // namespace io
} // namespace broccoli
