/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/curve/QuaternionSpline.hpp"
#include "gtest/gtest.h"
#include <fstream>
#include <iostream>
#include <string>

using namespace broccoli;
using namespace curve;

// Template function for plotting quaternion spline in 3D (using gnuplot)
template <class QuaternionCurveType>
void plotSpline3D(const QuaternionSpline<QuaternionCurveType>& spline, const double& stepSize, const std::string& fileName)
{
    // Writing gnuplot file
    std::ofstream gnuplotFile;
    gnuplotFile.open(fileName + ".gp");
    gnuplotFile << "set style line 1 linecolor rgb '#0065BD' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 2 linecolor rgb '#E37222' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 3 linecolor rgb '#A2AD00' linetype 1 linewidth 1\n";
    gnuplotFile << "\n";
    gnuplotFile << "splot '" << fileName << ".dat' using 2:3:4 with lines linestyle 1 title \"x-axis\"\n";
    gnuplotFile << "replot '" << fileName << ".dat' using 5:6:7 with lines linestyle 2 title \"y-axis\"\n";
    gnuplotFile << "replot '" << fileName << ".dat' using 8:9:10 with lines linestyle 3 title \"z-axis\"\n";
    gnuplotFile << "set view equal xyz\n";
    gnuplotFile.close();

    // Writing data file
    std::ofstream dataFile;
    dataFile.open(fileName + ".dat");
    int stepCount = (double)1.0 / stepSize + 1;
    for (int i = 0; i < stepCount; i++) {
        // Evaluate base function
        Eigen::Quaterniond point = spline.evaluate(i * stepSize, 0);

        dataFile << (double)i * stepSize;

        // Compute x-axis
        Eigen::Quaterniond xAxis = point * Eigen::Quaterniond(0, 1, 0, 0) * point.conjugate();
        dataFile << " " << xAxis.x() << " " << xAxis.y() << " " << xAxis.z();

        // Compute y-axis
        Eigen::Quaterniond yAxis = point * Eigen::Quaterniond(0, 0, 1, 0) * point.conjugate();
        dataFile << " " << yAxis.x() << " " << yAxis.y() << " " << yAxis.z();

        // Compute z-axis
        Eigen::Quaterniond zAxis = point * Eigen::Quaterniond(0, 0, 0, 1) * point.conjugate();
        dataFile << " " << zAxis.x() << " " << zAxis.y() << " " << zAxis.z();

        // Switch to next data set
        dataFile << "\n";
    }
    dataFile.close();
}

// Helpers to compare quaternions
// ------------------------------
static bool areQuaternionsEquivalent(const Eigen::Quaterniond& first, const Eigen::Quaterniond& second)
{
    // Orientations are equivalent if q1 = q2 OR q1 = -q2
    return (first.isApprox(second) || first.isApprox(Eigen::Quaterniond(-second.w(), -second.x(), -second.y(), -second.z())));
}
static bool areQuaternionsClose(const Eigen::Quaterniond& first, const Eigen::Quaterniond& second)
{
    // Compare dot-product (~angle between orientations)
    double dotProduct = fabs(acos(first.normalized().dot(second.normalized())));
    return dotProduct < 1e-5;
}

// Template function to compare analytic and numeric derivative
template <class QuaternionCurveType>
static bool compareAnalyticAndNumericDerivative(const QuaternionSpline<QuaternionCurveType>& spline, const unsigned int& derivationOrder)
{
    double dt = 0.0001;
    double maximumLocalError = 0;
    double maximumNorm = 0;

    // Pass through timesteps
    double t = dt;
    while (t <= 1.0 - dt) {
        // Compute analytic derivative
        Eigen::Quaterniond analyticDerivative = spline.evaluate(t, derivationOrder);

        // Compute numeric derivative dq/dt
        Eigen::Quaterniond leftSide = spline.evaluate(t - dt, derivationOrder - 1);
        Eigen::Quaterniond rightSide = spline.evaluate(t + dt, derivationOrder - 1);
        Eigen::Quaterniond numericDerivative;
        numericDerivative.coeffs() = (rightSide.coeffs() - leftSide.coeffs()) / (2 * dt);

        // Compute error
        Eigen::Quaterniond difference;
        difference.coeffs() = analyticDerivative.coeffs() - numericDerivative.coeffs();
        double localError = difference.norm();
        if (localError > maximumLocalError)
            maximumLocalError = localError;

        // Remember norm (for error estimation)
        double norm = analyticDerivative.norm();
        if (norm > maximumNorm)
            maximumNorm = norm;

        // Update time
        t = t + 0.001;
    }

    // Check for relative error (1%)
    double maximumAllowedLocalError = 1e-2;
    if (maximumNorm > 0)
        maximumAllowedLocalError = 0.01 * maximumNorm;
    if (maximumLocalError > maximumAllowedLocalError)
        return false;

    return true;
}

template <class CurveType>
void TestQuaternionSplineIsValid()
{
    // Initialize helpers
    QuaternionSplineResult result;

    // Initialize container
    QuaternionSpline<CurveType> spline;

    // Check no segments
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, QuaternionSplineResult::ERROR_INVALID_SPLINE_NO_SEGMENTS);

    // Check dimension mismatch
    spline.m_segments.push_back(CurveType());
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT);

    // Check negative proportions
    spline.m_segmentProportions.push_back(-1);
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE);

    // Check wrong sum of proportions
    spline.m_segmentProportions.clear();
    spline.m_segmentProportions.push_back(0.5);
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, QuaternionSplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM);

    // Check success
    spline.m_segmentProportions.clear();
    spline.m_segmentProportions.push_back(1);
    ASSERT_EQ(spline.isValid(&result), true);
    ASSERT_EQ(result, QuaternionSplineResult::SUCCESS);
}

//! Check removeEmptySegments() function
TEST(QuaternionSpline, removeEmptySegments)
{
    QuaternionSpline<QuaternionSLERPCurve> spline;
    spline.m_segments.resize(13);
    for (size_t i = 0; i < 13; i++)
        spline.m_segments[i].m_controlPoints[0] = Eigen::Quaterniond(i, 0, 0, 0);
    spline.m_segmentProportions.resize(13, 0.1);
    spline.m_segmentProportions[1] = 0;
    spline.m_segmentProportions[2] = 0;
    spline.m_segmentProportions[4] = 0;
    ASSERT_EQ(spline.removeEmptySegments(), 3);
    ASSERT_EQ(spline.m_segments.size(), 10);
    ASSERT_EQ(spline.m_segmentProportions.size(), 10);
    ASSERT_EQ(spline.m_segments[0].m_controlPoints[0].w(), 0);
    ASSERT_EQ(spline.m_segments[1].m_controlPoints[0].w(), 3);
    ASSERT_EQ(spline.m_segments[2].m_controlPoints[0].w(), 5);
    ASSERT_EQ(spline.m_segments[3].m_controlPoints[0].w(), 6);
    ASSERT_EQ(spline.m_segments[4].m_controlPoints[0].w(), 7);
    ASSERT_EQ(spline.m_segments[5].m_controlPoints[0].w(), 8);
    ASSERT_EQ(spline.m_segments[6].m_controlPoints[0].w(), 9);
    ASSERT_EQ(spline.m_segments[7].m_controlPoints[0].w(), 10);
    ASSERT_EQ(spline.m_segments[8].m_controlPoints[0].w(), 11);
    ASSERT_EQ(spline.m_segments[9].m_controlPoints[0].w(), 12);
}

//! Test string representation of types
TEST(QuaternionSpline, TypesToString)
{
    // Just check, if string is non-empty
    ASSERT_GT(QuaternionCurve::interpolationTypeString(QuaternionCurve::InterpolationType::SLERP).size(), 0);
    ASSERT_GT(quaternionSplineInterpolationMethodString(QuaternionSplineInterpolationMethod::SLERP).size(), 0);
    ASSERT_GT(quaternionSplineResultString(QuaternionSplineResult::SUCCESS).size(), 0);
}

//! Check isValid() function
TEST(QuaternionSpline, isValid)
{
    // Test slerp
    TestQuaternionSplineIsValid<QuaternionSLERPCurve>();
}

template <class CurveType>
void TestQuaternionSplineEncoding()
{
    // Initialize helpers
    QuaternionSpline<CurveType> spline;
    spline.m_segments.push_back(CurveType());
    io::encoding::CharacterStream stream;
    io::encoding::CharacterStreamSize streamSize = spline.encodeToXML(stream, 1, 4);
    ASSERT_EQ(streamSize > 0, true);
    ASSERT_EQ(streamSize, stream.size());
}

//! Check encoding
TEST(QuaternionSpline, encode)
{
    // Test slerp
    TestQuaternionSplineEncoding<QuaternionSLERPCurve>();
}

//! Check computation of quaternion curves
TEST(QuaternionSpline, QuaternionCurve)
{
    // Create control point quaternions
    Eigen::Quaterniond beginQuaternion(1, 0, 0, 0);
    Eigen::Quaterniond endQuaternion(1, 1, 0, 0);
    beginQuaternion.normalize();
    endQuaternion.normalize();

    // Setup curves
    QuaternionSLERPCurve slerpCurve(beginQuaternion, endQuaternion);

    // Create test quaternions
    Eigen::Quaterniond intermediateQuaternion(0.92388, 0.382683, 0, 0);

    // Check evaluation
    ASSERT_TRUE(areQuaternionsEquivalent(slerpCurve.evaluate(-0.1, 0), beginQuaternion));
    ASSERT_TRUE(areQuaternionsEquivalent(slerpCurve.evaluate(0.0, 0), beginQuaternion));
    ASSERT_TRUE(areQuaternionsClose(slerpCurve.evaluate(0.5, 0), intermediateQuaternion));
    ASSERT_TRUE(areQuaternionsEquivalent(slerpCurve.evaluate(1.0, 0), endQuaternion));
    ASSERT_TRUE(areQuaternionsEquivalent(slerpCurve.evaluate(1.1, 0), endQuaternion));

    // Check encoding
    io::encoding::CharacterStream stream;
    io::encoding::CharacterStreamSize streamSize = slerpCurve.encodeToXML(stream, 1, 4);
    ASSERT_TRUE(streamSize > 0);
    ASSERT_EQ(stream.size(), streamSize);

    // Check invalid parameters
    // ------------------------
    slerpCurve = QuaternionSLERPCurve(beginQuaternion, beginQuaternion);

    // Check evaluation
    ASSERT_TRUE(areQuaternionsEquivalent(slerpCurve.evaluate(0.5, 0), beginQuaternion));
    ASSERT_TRUE(areQuaternionsEquivalent(slerpCurve.evaluate(1.1, 1), Eigen::Quaterniond(0, 0, 0, 0)));
}

//! Check computation of quaternion splines
TEST(QuaternionSpline, QuaternionSpline)
{
    // Initialize helpers
    QuaternionSplineResult result;

    // Create control point quaternions
    std::vector<Eigen::Quaterniond, Eigen::aligned_allocator<Eigen::Quaterniond>> keyFrames;
    keyFrames.push_back(Eigen::Quaterniond(1, 0, 0, 0));
    keyFrames.push_back(Eigen::Quaterniond(1, 0, 0, 1));
    keyFrames.push_back(Eigen::Quaterniond(0.5, -0.5, 0.5, 0.5));
    keyFrames.push_back(Eigen::Quaterniond(0, -1, 0, 1));
    keyFrames.push_back(Eigen::Quaterniond(1, 0, 0, 0));
    for (size_t i = 0; i < keyFrames.size(); i++)
        keyFrames[i].normalize();

    // Setup spline
    QuaternionSpline<QuaternionSLERPCurve> spline;
    bool returnValue = spline.interpolate(keyFrames, nullptr /* empty -> use uniform proportions */, QuaternionSplineInterpolationMethod::SLERP, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, QuaternionSplineResult::SUCCESS);

    // Write plot files for visual debugging
    plotSpline3D<QuaternionSLERPCurve>(spline, 0.001, "QuaternionSLERPCurve3D");

    // Check derivatives numerically
    ASSERT_TRUE(compareAnalyticAndNumericDerivative<QuaternionSLERPCurve>(spline, 1));
    ASSERT_TRUE(compareAnalyticAndNumericDerivative<QuaternionSLERPCurve>(spline, 2));
    ASSERT_TRUE(compareAnalyticAndNumericDerivative<QuaternionSLERPCurve>(spline, 3));
}
