/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/curve/Spline.hpp"
#include "gtest/gtest.h"
#include <fstream>
#include <iostream>
#include <string>

using namespace broccoli;
using namespace curve;

// Template function for plotting spline (using gnuplot)
template <class CurveType>
void plotSpline(const Spline<CurveType>& spline, const double& stepSize, const unsigned int& maximumDerivation, const std::string& fileName)
{
    // Writing gnuplot file
    std::ofstream gnuplotFile;
    gnuplotFile.open(fileName + ".gp");
    gnuplotFile << "set style line 1 linecolor rgb '#0065BD' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 2 linecolor rgb '#E37222' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 3 linecolor rgb '#A2AD00' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 4 linecolor rgb '#5BC5F2' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 5 linecolor rgb '#E3828F' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 6 linecolor rgb '#7F7F7F' linetype 1 linewidth 1\n";
    gnuplotFile << "set style line 7 linecolor rgb '#000000' linetype 1 linewidth 1\n";
    gnuplotFile << "\n";
    gnuplotFile << "plot '" << fileName << ".dat' using 1:2 with lines linestyle 1 title \"Base function\"\n";
    for (unsigned int i = 1; i <= maximumDerivation; i++)
        gnuplotFile << "replot '" << fileName << ".dat' using 1:" << (2 + i) << " with lines linestyle " << (1 + i) << " title \"Derivative of order " << i << "\"\n";
    gnuplotFile.close();

    // Writing data file
    std::ofstream dataFile;
    dataFile.open(fileName + ".dat");
    int stepCount = (double)1.0 / stepSize + 1;
    for (int i = 0; i < stepCount; i++) {
        // Evaluate base function
        dataFile << (double)i * stepSize << " " << spline.evaluate(i * stepSize, 0);

        // Evaluate derivative
        for (unsigned int j = 1; j <= maximumDerivation; j++)
            dataFile << " " << spline.evaluate(i * stepSize, j);

        // Switch to next data set
        dataFile << "\n";
    }
    dataFile.close();
}

// Template function to compare analytic and numeric derivative
template <class CurveType>
bool compareDerivativeAnalyticNumeric(const Spline<CurveType>& spline, const unsigned int& derivationOrder)
{
    double dt = 0.0001;
    double maximumLocalError = 0;
    double maximumNorm = 0;

    // Pass through timesteps
    double t = dt;
    while (t <= 1.0 - dt) {
        // Compute analytic derivative
        double analyticDerivative = spline.evaluate(t, derivationOrder);

        // Compute numeric derivative
        double leftSide = spline.evaluate(t - dt, derivationOrder - 1);
        double rightSide = spline.evaluate(t + dt, derivationOrder - 1);
        double numericDerivative = (rightSide - leftSide) / (2.0 * dt);

        // Compute error
        double difference = analyticDerivative - numericDerivative;
        double localError = fabs(difference);
        if (localError > maximumLocalError)
            maximumLocalError = localError;

        // Remember norm (for error estimation)
        double norm = fabs(analyticDerivative);
        if (norm > maximumNorm)
            maximumNorm = norm;

        // Update time
        t = t + 0.001;
    }

    // Check for relative error (1%)
    double maximumAllowedLocalError = 1e-2;
    if (maximumNorm > 0)
        maximumAllowedLocalError = 0.01 * maximumNorm;
    if (maximumLocalError > maximumAllowedLocalError)
        return false;

    return true;
}

template <class CurveType>
void TestSplineIsValid()
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<CurveType> spline;

    // Check no segments
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, SplineResult::ERROR_INVALID_SPLINE_NO_SEGMENTS);

    // Check dimension mismatch
    spline.m_segments.push_back(CurveType());
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, SplineResult::ERROR_INVALID_SPLINE_PROPORTION_COUNT);

    // Check negative proportions
    spline.m_segmentProportions.push_back(-1);
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, SplineResult::ERROR_INVALID_SPLINE_PROPORTION_NEGATIVE);

    // Check wrong sum of proportions
    spline.m_segmentProportions.clear();
    spline.m_segmentProportions.push_back(0.5);
    ASSERT_EQ(spline.isValid(&result), false);
    ASSERT_EQ(result, SplineResult::ERROR_INVALID_SPLINE_PROPORTION_SUM);

    // Check success
    spline.m_segmentProportions.clear();
    spline.m_segmentProportions.push_back(1);
    ASSERT_EQ(spline.isValid(&result), true);
    ASSERT_EQ(result, SplineResult::SUCCESS);
}

//! Test string representation of types
TEST(Spline, TypesToString)
{
    // Just check, if string is non-empty
    ASSERT_GT(Curve::functionTypeString(Curve::FunctionType::POLYNOMIAL).size(), 0);
    ASSERT_GT(Trigonometric::typeString(Trigonometric::Type::SINE).size(), 0);
    ASSERT_GT(splineInterpolationMethodString(SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT).size(), 0);
    ASSERT_GT(splineResultString(SplineResult::SUCCESS).size(), 0);
}

//! Check isValid() function
TEST(Spline, isValid)
{
    // Test polynomials
    TestSplineIsValid<Polynomial<0>>();
    TestSplineIsValid<Polynomial<1>>();
    TestSplineIsValid<Polynomial<2>>();
    TestSplineIsValid<Polynomial<3>>();
    TestSplineIsValid<Polynomial<4>>();
    TestSplineIsValid<Polynomial<5>>();
    TestSplineIsValid<Polynomial<6>>();
    TestSplineIsValid<Polynomial<7>>();
    TestSplineIsValid<Polynomial<8>>();
    TestSplineIsValid<Polynomial<9>>();
    TestSplineIsValid<Polynomial<10>>();

    // Test trigonometrics
    TestSplineIsValid<Trigonometric>();

    // Test exponential
    TestSplineIsValid<Exponential>();
}

//! Check removeEmptySegments() function
TEST(Spline, removeEmptySegments)
{
    Spline<Polynomial<1>> spline;
    spline.m_segments.resize(13);
    for (size_t i = 0; i < 13; i++)
        spline.m_segments[i].m_coefficients[0] = i;
    spline.m_segmentProportions.resize(13, 0.1);
    spline.m_segmentProportions[1] = 0;
    spline.m_segmentProportions[2] = 0;
    spline.m_segmentProportions[4] = 0;
    ASSERT_EQ(spline.removeEmptySegments(), 3);
    ASSERT_EQ(spline.m_segments.size(), 10);
    ASSERT_EQ(spline.m_segmentProportions.size(), 10);
    ASSERT_EQ(spline.m_segments[0].m_coefficients[0], 0);
    ASSERT_EQ(spline.m_segments[1].m_coefficients[0], 3);
    ASSERT_EQ(spline.m_segments[2].m_coefficients[0], 5);
    ASSERT_EQ(spline.m_segments[3].m_coefficients[0], 6);
    ASSERT_EQ(spline.m_segments[4].m_coefficients[0], 7);
    ASSERT_EQ(spline.m_segments[5].m_coefficients[0], 8);
    ASSERT_EQ(spline.m_segments[6].m_coefficients[0], 9);
    ASSERT_EQ(spline.m_segments[7].m_coefficients[0], 10);
    ASSERT_EQ(spline.m_segments[8].m_coefficients[0], 11);
    ASSERT_EQ(spline.m_segments[9].m_coefficients[0], 12);
}

template <class CurveType>
void TestSplineEncoding()
{
    // Initialize helpers
    Spline<CurveType> spline;
    spline.m_segments.push_back(CurveType());
    io::encoding::CharacterStream stream;
    io::encoding::CharacterStreamSize streamSize = spline.encodeToXML(stream, 1, 4);
    ASSERT_EQ(streamSize > 0, true);
    ASSERT_EQ(streamSize, stream.size());
}

//! Check encoding
TEST(Spline, encode)
{
    // Test polynomials
    TestSplineEncoding<Polynomial<0>>();
    TestSplineEncoding<Polynomial<1>>();
    TestSplineEncoding<Polynomial<2>>();
    TestSplineEncoding<Polynomial<3>>();
    TestSplineEncoding<Polynomial<4>>();
    TestSplineEncoding<Polynomial<5>>();
    TestSplineEncoding<Polynomial<6>>();
    TestSplineEncoding<Polynomial<7>>();
    TestSplineEncoding<Polynomial<8>>();
    TestSplineEncoding<Polynomial<9>>();
    TestSplineEncoding<Polynomial<10>>();

    // Test trigonometrics
    TestSplineEncoding<Trigonometric>();

    // Test exponential
    TestSplineEncoding<Exponential>();
}

//! Check construction and interpolation of polynomial of order zero
TEST(Spline, PolynomialOfOrderZero)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<0>> spline;

    // Check type:
    ASSERT_EQ(Polynomial<0>::functionType(), Curve::FunctionType::POLYNOMIAL);

    // Step 1: test with non-uniform (custom) proportions
    // --------------------------------------------------
    // Setup parameter set
    std::vector<double> parameters;
    std::vector<double> proportions;
    parameters.resize(5);
    proportions.resize(5);

    // Function values
    parameters[0] = -2;
    parameters[1] = -1;
    parameters[2] = 0;
    parameters[3] = 1;
    parameters[4] = 2;

    // Proportions
    proportions[0] = 0.2;
    proportions[1] = 0.2;
    proportions[2] = 0.2;
    proportions[3] = 0.2;
    proportions[4] = 0.2;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size());
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Check function values
    ASSERT_TRUE(spline.evaluate(0.0000, 0) == -2);
    ASSERT_TRUE(spline.evaluate(0.1000, 0) == -2);
    ASSERT_TRUE(spline.evaluate(0.1999, 0) == -2);
    ASSERT_TRUE(spline.evaluate(0.2001, 0) == -1);
    ASSERT_TRUE(spline.evaluate(0.3999, 0) == -1);
    ASSERT_TRUE(spline.evaluate(0.4001, 0) == 0);
    ASSERT_TRUE(spline.evaluate(0.5999, 0) == 0);
    ASSERT_TRUE(spline.evaluate(0.6001, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.7999, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.8001, 0) == 2);
    ASSERT_TRUE(spline.evaluate(1.0000, 0) == 2);

    // Check derivatives
    ASSERT_TRUE(spline.evaluate(0.1, 1) == 0);
    ASSERT_TRUE(spline.evaluate(0.1, 2) == 0);
    ASSERT_TRUE(spline.evaluate(0.1, 3) == 0);
    ASSERT_TRUE(spline.evaluate(0.1, 4) == 0);

    // Step 2: test with uniform proportions
    // -------------------------------------
    // Setup parameter set
    parameters.resize(5);
    proportions.clear();

    // Function values
    parameters[0] = -2;
    parameters[1] = -1;
    parameters[2] = 0;
    parameters[3] = 1;
    parameters[4] = 2;

    // Trigger interpolation (wrong method)
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_LINEAR, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);

    // Trigger interpolation (correct method)
    returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size());
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<0>>(spline, 0.001, 1, "PolynomialOfOrderZero");

    // Check function values
    ASSERT_TRUE(spline.evaluate(0.0000, 0) == -2);
    ASSERT_TRUE(spline.evaluate(0.1000, 0) == -2);
    ASSERT_TRUE(spline.evaluate(0.1999, 0) == -2);
    ASSERT_TRUE(spline.evaluate(0.2001, 0) == -1);
    ASSERT_TRUE(spline.evaluate(0.3999, 0) == -1);
    ASSERT_TRUE(spline.evaluate(0.4001, 0) == 0);
    ASSERT_TRUE(spline.evaluate(0.5999, 0) == 0);
    ASSERT_TRUE(spline.evaluate(0.6001, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.7999, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.8001, 0) == 2);
    ASSERT_TRUE(spline.evaluate(1.0000, 0) == 2);

    // Check derivatives
    ASSERT_TRUE(spline.evaluate(0.1, 1) == 0);
    ASSERT_TRUE(spline.evaluate(0.1, 2) == 0);
    ASSERT_TRUE(spline.evaluate(0.1, 3) == 0);
    ASSERT_TRUE(spline.evaluate(0.1, 4) == 0);

    // Step 3: compare analytic and numeric derivatives
    // ------------------------------------------------
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<0>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<0>>(spline, 2));
}

//! Check construction and interpolation of polynomial of order one
TEST(Spline, PolynomialOfOrderOne)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<1>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(3);
    parameters[0] = 0;
    parameters[1] = 1;
    parameters[2] = -1;

    // Trigger interpolation (wrong method)
    bool returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);

    // Trigger interpolation (correct method)
    returnValue = spline.interpolate(parameters, nullptr /* empty -> use uniform proportions */, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_LINEAR, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 1);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<1>>(spline, 0.001, 2, "PolynomialOfOrderOne");

    // Check function values
    ASSERT_TRUE(spline.evaluate(0.00, 0) == 0);
    ASSERT_TRUE(spline.evaluate(0.25, 0) == 0.5);
    ASSERT_TRUE(spline.evaluate(0.50, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.75, 0) == 0);
    ASSERT_TRUE(spline.evaluate(1.00, 0) == -1);

    // Check derivatives
    ASSERT_TRUE(spline.evaluate(0.25, 1) == 2);
    ASSERT_TRUE(spline.evaluate(0.75, 1) == -4);
    ASSERT_TRUE(spline.evaluate(0.25, 2) == 0);
    ASSERT_TRUE(spline.evaluate(0.75, 2) == 0);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<1>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<1>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<1>>(spline, 3));
}

//! Check construction and interpolation of polynomial of order three (with given first order derivatives)
TEST(Spline, PolynomialOfOrderThreeFirstDerivatives)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<3>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(10, 0);
    parameters[0] = 0;
    parameters[1] = 1;
    parameters[2] = 1;
    parameters[3] = 1;
    parameters[4] = 0;
    parameters[5] = 0;
    parameters[6] = 10;
    parameters[7] = 10;
    parameters[8] = 10;
    parameters[9] = 0;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, nullptr /* empty -> use uniform proportions */, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_FIRST_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() / 2 - 1);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<3>>(spline, 0.001, 2, "PolynomialOfOrderThreeFirstDerivatives");

    // Check function values
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 0) - 0) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 0) - 1) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 0) - 1) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 0) - 1) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 0) - 0) < 1e-10);

    // Check derivatives
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 1) - 0) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 1) - 10) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 1) - 10) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 1) - 10) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 1) - 0) < 1e-10);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 4));
}

//! Check construction and interpolation of polynomial of order three (with given second order derivatives)
TEST(Spline, PolynomialOfOrderThreeSecondDerivatives)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<3>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(10, 0);
    parameters[0] = 0;
    parameters[1] = 1;
    parameters[2] = 1;
    parameters[3] = 1;
    parameters[4] = 0;
    parameters[5] = 0;
    parameters[6] = 10;
    parameters[7] = 10;
    parameters[8] = 10;
    parameters[9] = 0;

    // Trigger interpolation (wrong method)
    bool returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);
    returnValue = spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, false);
    ASSERT_EQ(result, SplineResult::ERROR_DEGREE_TOO_LOW);

    // Trigger interpolation (correct method)
    returnValue = spline.interpolate(parameters, nullptr /* empty -> use uniform proportions */, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() / 2 - 1);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<3>>(spline, 0.001, 2, "PolynomialOfOrderThreeSecondDerivatives");

    // Check function values
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 0) - 0) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 0) - 1) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 0) - 1) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 0) - 1) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 0) - 0) < 1e-10);

    // Check derivatives
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 2) - 0) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 2) - 10) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 2) - 10) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 2) - 10) < 1e-10);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 2) - 0) < 1e-10);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 4));
}

//! Check construction and interpolation of polynomial of order five (given first and second order derivatives)
TEST(Spline, PolynomialOfOrderFiveFirstAndSecondDerivatives)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<5>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(15, 0);
    parameters[0] = 0;
    parameters[1] = 1;
    parameters[2] = 1;
    parameters[3] = 1;
    parameters[4] = 0;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, nullptr /* empty -> use uniform proportions */, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() / 3 - 1);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<5>>(spline, 0.001, 2, "PolynomialOfOrderFiveFirstAndSecondDerivatives");

    // Check function values
    ASSERT_TRUE(spline.evaluate(0.00, 0) == 0);
    ASSERT_TRUE(spline.evaluate(0.125, 0) == 0.5);
    ASSERT_TRUE(spline.evaluate(0.25, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.50, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.75, 0) == 1);
    ASSERT_TRUE(spline.evaluate(0.875, 0) == 0.5);
    ASSERT_TRUE(spline.evaluate(1.00, 0) == 0);

    // Check derivatives
    ASSERT_TRUE(spline.evaluate(0.00, 1) == 0);
    ASSERT_TRUE(spline.evaluate(0.25, 1) == 0);
    ASSERT_TRUE(spline.evaluate(0.50, 1) == 0);
    ASSERT_TRUE(spline.evaluate(0.75, 1) == 0);
    ASSERT_TRUE(spline.evaluate(1.00, 1) == 0);
    ASSERT_TRUE(spline.evaluate(0.00, 2) == 0);
    ASSERT_TRUE(spline.evaluate(0.25, 2) == 0);
    ASSERT_TRUE(spline.evaluate(0.50, 2) == 0);
    ASSERT_TRUE(spline.evaluate(0.75, 2) == 0);
    ASSERT_TRUE(spline.evaluate(1.00, 2) == 0);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 4));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 5));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 6));
}

//! Check construction and interpolation of smooth cubic spline with only one segment (special case)
TEST(Spline, SmoothCubicSplineSingleSegment)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<3>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(4, 0);
    parameters[0] = 1.0;
    parameters[1] = 2.0;
    parameters[2] = 0.1;
    parameters[3] = 0.2;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, nullptr /* empty -> use uniform proportions */, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 3);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<3>>(spline, 0.001, 2, "SmoothCubicSplineSingleSegment");

    // Check interpolation
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 0) - parameters[0]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 0) - parameters[1]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 2) - parameters[2]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 2) - parameters[3]) < 1e-6);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 4));
}

//! Check construction and interpolation of smooth cubic spline with only two segments (special case)
TEST(Spline, SmoothCubicSplineTwoSegments)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<3>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(5, 0);
    parameters[0] = 0.5;
    parameters[1] = 1.5;
    parameters[2] = 2.0;
    parameters[3] = 0.1;
    parameters[4] = 0.2;
    std::vector<double> proportions;
    proportions.resize(2);
    proportions[0] = 0.2;
    proportions[1] = 0.8;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 3);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<3>>(spline, 0.001, 3, "SmoothCubicSplineTwoSegments");

    // Check interpolation
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 0) - parameters[0]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.2, 0) - parameters[1]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 0) - parameters[2]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 2) - parameters[3]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 2) - parameters[4]) < 1e-6);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 4));
}

//! Check construction and interpolation of smooth cubic spline
TEST(Spline, SmoothCubicSpline)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<3>> spline;

    // Setup parameter set
    std::vector<double> proportions;
    std::vector<double> parameters;
    parameters.resize(7, 0);
    parameters[0] = 0.0;
    parameters[1] = 1.0;
    parameters[2] = 0.5;
    parameters[3] = 2.0;
    parameters[4] = 3.0;
    parameters[5] = 0.1;
    parameters[6] = 0.2;
    proportions.resize(4, 0);
    proportions[0] = 0.1;
    proportions[1] = 0.2;
    proportions[2] = 0.3;
    proportions[3] = 0.4;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_CUBIC_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 3);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<3>>(spline, 0.001, 2, "SmoothCubicSpline");

    // Check interpolation
    double splinePosition = 0;
    for (size_t i = 0; i < parameters.size() - 2; i++) {
        ASSERT_TRUE(fabs(spline.evaluate(splinePosition, 0) - parameters[i]) < 1e-6);
        if (i < proportions.size())
            splinePosition += proportions[i];
    }
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 2) - parameters[5]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 2) - parameters[6]) < 1e-6);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<3>>(spline, 4));
}

//! Check construction and interpolation of smooth quintic spline with only one segment (special case)
TEST(Spline, SmoothQuinticSplineSingleSegment)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<5>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(6, 0);
    parameters[0] = 1.0;
    parameters[1] = 2.0;
    parameters[2] = 3.0;
    parameters[3] = 4.0;
    parameters[4] = 0.1;
    parameters[5] = 0.2;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, nullptr /* empty -> use uniform proportions */, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 5);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<5>>(spline, 0.001, 5, "SmoothQuinticSplineSingleSegment");

    // Check interpolation
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 0) - parameters[0]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 0) - parameters[1]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 1) - parameters[2]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 1) - parameters[3]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 2) - parameters[4]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 2) - parameters[5]) < 1e-6);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 4));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 5));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 6));
}

//! Check construction and interpolation of smooth quintic spline with only two segments (special case)
TEST(Spline, SmoothQuinticSplineTwoSegments)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<5>> spline;

    // Setup parameter set
    std::vector<double> parameters;
    parameters.resize(7, 0);
    parameters[0] = 0.0;
    parameters[1] = 1.0;
    parameters[2] = 0.0;
    parameters[3] = 0.0;
    parameters[4] = 0.0;
    parameters[5] = 0.0;
    parameters[6] = 0.0;
    std::vector<double> proportions;
    proportions.resize(2);
    proportions[0] = 0.2;
    proportions[1] = 0.8;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 5);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<5>>(spline, 0.001, 5, "SmoothQuinticSplineTwoSegments");

    // Check interpolation
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 0) - parameters[0]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.2, 0) - parameters[1]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 0) - parameters[2]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 1) - parameters[3]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 1) - parameters[4]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 2) - parameters[5]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 2) - parameters[6]) < 1e-6);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 4));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 5));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 6));
}

//! Check construction and interpolation of smooth quintic spline
TEST(Spline, SmoothQuinticSpline)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Polynomial<5>> spline;

    // Setup parameter set
    std::vector<double> proportions;
    std::vector<double> parameters;
    parameters.resize(9, 0);
    parameters[0] = 0.0;
    parameters[1] = 1.0;
    parameters[2] = 0.5;
    parameters[3] = 2.0;
    parameters[4] = 3.0;
    parameters[5] = 0.1;
    parameters[6] = 0.2;
    parameters[7] = 0.3;
    parameters[8] = 0.4;
    proportions.resize(4, 0);
    proportions[0] = 0.1;
    proportions[1] = 0.2;
    proportions[2] = 0.3;
    proportions[3] = 0.4;

    // Trigger interpolation
    bool returnValue = spline.interpolate(parameters, &proportions, SplineInterpolationMethod::POLYNOMIAL_SMOOTH_QUINTIC_FIRST_AND_SECOND_DERIVATIVES, &result);
    ASSERT_EQ(returnValue, true);
    ASSERT_EQ(result, SplineResult::SUCCESS);

    // Check segment count
    ASSERT_TRUE(spline.m_segments.size() == parameters.size() - 5);
    ASSERT_TRUE(spline.m_segmentProportions.size() == spline.m_segments.size());

    // Write plot file for visual debugging
    plotSpline<Polynomial<5>>(spline, 0.001, 5, "SmoothQuinticSpline");

    // Check interpolation
    double splinePosition = 0;
    for (size_t i = 0; i < parameters.size() - 4; i++) {
        ASSERT_TRUE(fabs(spline.evaluate(splinePosition, 0) - parameters[i]) < 1e-6);
        if (i < proportions.size())
            splinePosition += proportions[i];
    }
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 1) - parameters[5]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 1) - parameters[6]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(0.0, 2) - parameters[7]) < 1e-6);
    ASSERT_TRUE(fabs(spline.evaluate(1.0, 2) - parameters[8]) < 1e-6);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 3));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 4));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 5));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Polynomial<5>>(spline, 6));
}

//! Check construction and evaluation of basic sine function
TEST(Spline, TrigonometricSine)
{
    // Initialize container
    Spline<Trigonometric> spline;

    // Check type:
    ASSERT_EQ(Trigonometric::functionType(), Curve::FunctionType::TRIGONOMETRIC);

    // Setup segments
    // f(x) = c[0] + c[1]*sin(c[2]*x + c[3])
    spline.m_segments.resize(1);
    spline.m_segments[0].m_type = Trigonometric::Type::SINE;
    spline.m_segments[0].m_coefficients[0] = 0; // Offset
    spline.m_segments[0].m_coefficients[1] = 1; // Amplitude
    spline.m_segments[0].m_coefficients[2] = 2.0 * M_PI; // Streching
    spline.m_segments[0].m_coefficients[3] = 2.0 * M_PI; // Phase shift

    // Setup proportions
    spline.m_segmentProportions.resize(1, 1.0);

    // Write plot file for visual debugging
    plotSpline<Trigonometric>(spline, 0.001, 2, "TrigonometricSine");

    // Check function values
    double tol = 1e-10;
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 0) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 0) - 1.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 0) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 0) + 1.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 0) - 0.0) < tol);

    // Check derivatives
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 1) - 2.0 * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 1) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 1) + 2.0 * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 1) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 1) - 2.0 * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 2) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 2) + 4.0 * M_PI * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 2) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 2) - 4.0 * M_PI * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 2) - 0.0) < tol);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 3));
}

//! Check construction and evaluation of basic cosine function
TEST(Spline, TrigonometricCosine)
{
    // Initialize container
    Spline<Trigonometric> spline;

    // Check type:
    ASSERT_EQ(Trigonometric::functionType(), Curve::FunctionType::TRIGONOMETRIC);

    // Setup segments
    // f(x) = c[0] + c[1]*cos(c[2]*x + c[3])
    spline.m_segments.resize(1);
    spline.m_segments[0].m_type = Trigonometric::Type::COSINE;
    spline.m_segments[0].m_coefficients[0] = 0; // Offset
    spline.m_segments[0].m_coefficients[1] = 1; // Amplitude
    spline.m_segments[0].m_coefficients[2] = 2.0 * M_PI; // Streching
    spline.m_segments[0].m_coefficients[3] = 2.0 * M_PI; // Phase shift

    // Setup proportions
    spline.m_segmentProportions.resize(1, 1.0);

    // Write plot file for visual debugging
    plotSpline<Trigonometric>(spline, 0.001, 2, "TrigonometricCosine");

    // Check function values
    double tol = 1e-10;
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 0) - 1.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 0) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 0) + 1.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 0) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 0) - 1.0) < tol);

    // Check derivatives
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 1) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 1) + 2.0 * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 1) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 1) - 2.0 * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 1) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 2) + 4.0 * M_PI * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 2) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 2) - 4.0 * M_PI * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 2) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 2) + 4.0 * M_PI * M_PI) < tol);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 3));
}

//! Check construction and evaluation of basic tangent function
TEST(Spline, TrigonometricTangent)
{
    // Initialize container
    Spline<Trigonometric> spline;

    // Check type:
    ASSERT_EQ(Trigonometric::functionType(), Curve::FunctionType::TRIGONOMETRIC);

    // Setup segments
    // f(x) = c[0] + c[1]*tan(c[2]*x + c[3])
    spline.m_segments.resize(1);
    spline.m_segments[0].m_type = Trigonometric::Type::TANGENT;
    spline.m_segments[0].m_coefficients[0] = 0; // Offset
    spline.m_segments[0].m_coefficients[1] = 1; // Amplitude
    spline.m_segments[0].m_coefficients[2] = M_PI / 4.0; // Streching
    spline.m_segments[0].m_coefficients[3] = 0; // Phase shift

    // Setup proportions
    spline.m_segmentProportions.resize(1, 1.0);

    // Write plot file for visual debugging
    plotSpline<Trigonometric>(spline, 0.001, 1, "TrigonometricTangent");

    // Check function values
    double tol = 1e-10;
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 0) - 0.0) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 0) - tan(M_PI / 4.0 * 0.5)) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 0) - tan(M_PI / 4.0 * 1.0)) < tol);

    // Check derivatives
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 1) - 1.0 / 4.0 * M_PI) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 1) - M_PI / 4.0 * (1.0 + tan(M_PI / 8.0) * tan(M_PI / 8.0))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 1) - 1.0 / 2.0 * M_PI) < tol);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Trigonometric>(spline, 1));
}

//! Check construction and evaluation of trigonometric functions (in combination)
TEST(Spline, TrigonometricCombination)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Trigonometric> spline;

    // NOT IMPLEMENTED: interpolation
    std::vector<double> parameters;
    ASSERT_TRUE(spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT, &result) == false);
    ASSERT_TRUE(result == SplineResult::ERROR_NOTIMPLEMENTED);

    // Setup segments
    // f(x) = c[0] + c[1]*func(c[2]*x + c[3])
    spline.m_segments.resize(3);
    spline.m_segments[0].m_type = Trigonometric::Type::SINE;
    spline.m_segments[0].m_coefficients[0] = 1; // Offset
    spline.m_segments[0].m_coefficients[1] = 2; // Amplitude
    spline.m_segments[0].m_coefficients[2] = 2; // Streching
    spline.m_segments[0].m_coefficients[3] = M_PI_2; // Phase shift
    spline.m_segments[1].m_type = Trigonometric::Type::COSINE;
    spline.m_segments[1].m_coefficients[0] = 1; // Offset
    spline.m_segments[1].m_coefficients[1] = 2; // Amplitude
    spline.m_segments[1].m_coefficients[2] = 2; // Streching
    spline.m_segments[1].m_coefficients[3] = M_PI_2; // Phase shift
    spline.m_segments[2].m_type = Trigonometric::Type::TANGENT;
    spline.m_segments[2].m_coefficients[0] = 1; // Offset
    spline.m_segments[2].m_coefficients[1] = 2; // Amplitude
    spline.m_segments[2].m_coefficients[2] = 2; // Streching
    spline.m_segments[2].m_coefficients[3] = 0; // Phase shift

    // Setup proportions
    spline.m_segmentProportions.resize(3, 1.0 / 3.0);

    // Write plot file for visual debugging
    plotSpline<Trigonometric>(spline, 0.001, 0, "TrigonometricCombination");
}

//! Check construction and evaluation of exponential functions
TEST(Spline, ExponentialFunction)
{
    // Initialize helpers
    SplineResult result;

    // Initialize container
    Spline<Exponential> spline;

    // NOT IMPLEMENTED: interpolation
    std::vector<double> parameters;
    ASSERT_TRUE(spline.interpolate(parameters, nullptr, SplineInterpolationMethod::POLYNOMIAL_PIECEWISE_CONSTANT, &result) == false);
    ASSERT_TRUE(result == SplineResult::ERROR_NOTIMPLEMENTED);

    // Check type:
    ASSERT_EQ(Exponential::functionType(), Curve::FunctionType::EXPONENTIAL);

    // Setup segments
    // f(x) = c[0] + c[1]*exp(c[2]*x)
    spline.m_segments.resize(1);
    spline.m_segments[0].m_coefficients[0] = 1; // Offset
    spline.m_segments[0].m_coefficients[1] = 2; // Scaling
    spline.m_segments[0].m_coefficients[2] = 3; // Streching

    // Setup proportions
    spline.m_segmentProportions.resize(1, 1.0);

    // Write plot file for visual debugging
    plotSpline<Exponential>(spline, 0.001, 1, "ExponentialFunction");

    // Get coefficients
    double c0 = spline.m_segments[0].m_coefficients[0];
    double c1 = spline.m_segments[0].m_coefficients[1];
    double c2 = spline.m_segments[0].m_coefficients[2];

    // Check function values
    double tol = 1e-10;
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 0) - (c0 + c1 * exp(c2 * 0.00))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 0) - (c0 + c1 * exp(c2 * 0.25))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 0) - (c0 + c1 * exp(c2 * 0.50))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 0) - (c0 + c1 * exp(c2 * 0.75))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 0) - (c0 + c1 * exp(c2 * 1.00))) < tol);

    // Check derivatives
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 1) - (c1 * c2 * exp(c2 * 0.00))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 1) - (c1 * c2 * exp(c2 * 0.25))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 1) - (c1 * c2 * exp(c2 * 0.50))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 1) - (c1 * c2 * exp(c2 * 0.75))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 1) - (c1 * c2 * exp(c2 * 1.00))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.00, 2) - (c1 * c2 * c2 * exp(c2 * 0.00))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.25, 2) - (c1 * c2 * c2 * exp(c2 * 0.25))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.50, 2) - (c1 * c2 * c2 * exp(c2 * 0.50))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(0.75, 2) - (c1 * c2 * c2 * exp(c2 * 0.75))) < tol);
    ASSERT_TRUE(fabs(spline.evaluate(1.00, 2) - (c1 * c2 * c2 * exp(c2 * 1.00))) < tol);

    // Compare analytic and numeric derivatives
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Exponential>(spline, 1));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Exponential>(spline, 2));
    ASSERT_TRUE(compareDerivativeAnalyticNumeric<Exponential>(spline, 3));
}
