/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/io/Joystick.hpp"
#include "gtest/gtest.h"

using namespace broccoli;
using namespace io;

#ifdef HAVE_SDL2

//! Test string representation of types
TEST(Joystick, TypesToString)
{
    // Just check, if string is non-empty
    JoystickAction testAction;
    JoystickEvent testEvent;
    JoystickInformation testInformation;
    JoystickSignal testSignal;
    ASSERT_GT(testAction.typeString().size(), 0);
    ASSERT_GT(testEvent.typeString().size(), 0);
    ASSERT_GT(testInformation.powerLevelString().size(), 0);
    ASSERT_GT(testSignal.typeString().size(), 0);
    ASSERT_GT(testSignal.messageString().size(), 0);
    ASSERT_GT(JoystickState::hatValueString(SDL_HAT_LEFTUP).size(), 0);
}

//! Test setters and getters
TEST(Joystick, SettersAndGetters)
{
    // Create instance
    Joystick joystick(0, "TestDevice");

    // Setters and getters
    std::string desiredDeviceName = "A";
    joystick.setDesiredDeviceName(desiredDeviceName);
    ASSERT_EQ(desiredDeviceName, joystick.desiredDeviceName());
    unsigned int motionThreshold = 123;
    joystick.setMotionThreshold(motionThreshold);
    ASSERT_EQ(motionThreshold, joystick.motionThreshold());

    // Getters only
    bool isSDLInitialized = joystick.isSDLInitialized();
    ASSERT_EQ(isSDLInitialized, false);
    bool isConnected = joystick.isConnected();
    ASSERT_EQ(isConnected, false);
    std::vector<JoystickInformation> availableJoysticks = joystick.availableJoysticks();
    (void)availableJoysticks;
    JoystickInformation activeJoystick = joystick.activeJoystick();
    (void)activeJoystick;
    JoystickState initialState = joystick.initialState();
    (void)initialState;
    JoystickState currentState = joystick.currentState();
    (void)currentState;
}

//! Check basic creation and handling of joystick instance
TEST(Joystick, BasicHandling)
{
    // Create joystick instace
    Joystick joystick(0, "TestDevice");

    // Copy constructor
    Joystick joystickCopy(joystick);
    ASSERT_TRUE(joystickCopy.desiredDeviceName() == joystick.desiredDeviceName());

    // Assignment operator
    joystickCopy.setDesiredDeviceName("abc");
    joystickCopy = joystick;
    ASSERT_TRUE(joystickCopy.desiredDeviceName() == joystick.desiredDeviceName());

    // Initialize joystick module
    joystick.initialize();

    // Check, if SDL was initialized successfully
    ASSERT_EQ(joystick.isSDLInitialized(), true);

    // Wait until background joystick of joystick was executed at least once
    while (joystick.executionCalls() <= 1) {
        // Wait...
    }

    // Get list of available devices
    std::vector<JoystickInformation> availableJoysticks = joystick.availableJoysticks();

    // Output device list
    std::cout << "available joysticks:" << std::endl;
    for (size_t i = 0; i < availableJoysticks.size(); i++) {
        std::cout << "#" << i << ": Name: " << availableJoysticks[i].m_name << " | GUID: " << availableJoysticks[i].m_GUIDstring << std::endl;
    }
    if (availableJoysticks.size() == 0)
        std::cout << "none!" << std::endl;

    // Deinitialize joystick module and check for proper shutdown
    ASSERT_EQ(joystick.deInitialize(), true);
}

#else
#warning("Skipped Joystick test as SDL2 was not found!")
#endif
