/*
 * This file is part of broccoli.
 * Copyright (C) 2019 Chair of Applied Mechanics, Technical University of Munich
 * https://www.amm.mw.tum.de/
 */

#include "broccoli/io/NetworkSocket.hpp"
#include "broccoli/io/SerializableDataString.hpp"
#include "gtest/gtest.h"

using namespace broccoli;
using namespace io;

//! Test of networksocket communication for different socket types
void testSocketCommunication(const NetworkSocketType& socketTypeA, const NetworkSocketType& socketTypeB)
{
    // Create sockets
    NetworkSocket<SerializableDataString, SerializableDataString> socketA(socketTypeA, "socketA", "localhost", 5006, 5005, true, 0, 0, true, 10);
    NetworkSocket<SerializableDataString, SerializableDataString> socketB(socketTypeB, "socketB", "localhost", 5005, 5006, true, 0, 0, true, 10);

    // Copy constructor
    NetworkSocket<SerializableDataString, SerializableDataString> socketCopy(socketA);
    ASSERT_TRUE(socketCopy.name() == socketA.name());

    // Assignment operator
    socketCopy = socketB;
    ASSERT_TRUE(socketCopy.name() == socketB.name());

    // Redefine heartbeat timeouts
    socketA.setHeartBeatSendInterval(core::Time(0.001));
    socketB.setHeartBeatSendInterval(core::Time(0.001));
    socketA.setHeartBeatTimeout(core::Time(0.01));
    socketB.setHeartBeatTimeout(core::Time(0.01));

    // Setup
    // -----
    // Initialize sockets
    ASSERT_EQ(socketA.initialize(), true);
    ASSERT_EQ(socketB.initialize(), true);

    // Wait to allow establishing connection
    core::Time::sleep(0.01);

    // Check connection status
    ASSERT_EQ(socketA.isConnected(), true);
    ASSERT_EQ(socketB.isConnected(), true);

    // Heartbeat detection
    // -------------------
    // Wait to check for heartbeat timeout
    core::Time::sleep(0.03);
    ASSERT_EQ(socketA.isConnected(), true);
    ASSERT_EQ(socketB.isConnected(), true);

    // Check connection status
    ASSERT_EQ(socketA.isConnected(), true);
    ASSERT_EQ(socketB.isConnected(), true);

    // Sending and receiving
    // ---------------------
    // Add message to send buffer
    SerializableDataString testMessageToSend("test");
    ASSERT_EQ(socketA.m_objectsToSend.push(testMessageToSend), 0);

    // Wait to give sending/receiving some time...
    core::Time::sleep(0.02);

    // Try to get message
    SerializableDataString testMessageToReceive;
    ASSERT_EQ(socketB.m_receivedObjects.pop(testMessageToReceive), 1);

    // Compare messages
    ASSERT_EQ(testMessageToReceive, testMessageToSend);

    // Shutdown
    // --------
    // De-initialize one socket to simulate disconnection
    ASSERT_EQ(socketA.deInitialize(), true);

    // Wait to check for heartbeat timeout
    core::Time::sleep(0.03);
    ASSERT_EQ(socketA.isConnected(), false);
    ASSERT_EQ(socketB.isConnected(), false);

    // De-initialize other socket and check for proper shutdown
    ASSERT_EQ(socketB.deInitialize(), true);
}

//! Test string representation of types
TEST(NetworkSocket, TypesToString)
{
    // Just check, if string is non-empty
    NetworkSocketSignal testSignal;
    ASSERT_GT(networkSocketTypeString(NetworkSocketType::TCP_SERVER).size(), 0);
    ASSERT_GT(testSignal.typeString().size(), 0);
    ASSERT_GT(testSignal.messageString().size(), 0);
}

//! Check implementation of UDP communication
TEST(NetworkSocket, UDPCommunication)
{
    testSocketCommunication(NetworkSocketType::UDP, NetworkSocketType::UDP);
}

//! Check implementation of TCP communication
TEST(NetworkSocket, TCPCommunication)
{
    testSocketCommunication(NetworkSocketType::TCP_SERVER, NetworkSocketType::TCP_CLIENT);
}

//! Test setters and getters
TEST(NetworkSocket, SettersAndGetters)
{
    // Create instance
    NetworkSocket<SerializableDataString, SerializableDataString> socket(NetworkSocketType::UDP, "socket", "localhost", 5006, 5005, true, 0, 0, true, 10);

    // Setters and getters
    core::Time connectionAttemptInterval = core::Time(1);
    socket.setConnectionAttemptInterval(connectionAttemptInterval);
    ASSERT_EQ(connectionAttemptInterval, socket.connectionAttemptInterval());
    bool sendHeartBeats = true;
    socket.setSendHeartBeats(sendHeartBeats);
    ASSERT_EQ(sendHeartBeats, socket.sendHeartBeats());
    NetworkSocket<SerializableDataString, SerializableDataString>::DataStream heartBeatSendCode = { 0x1A };
    socket.setHeartBeatSendCode(heartBeatSendCode);
    ASSERT_EQ(heartBeatSendCode, socket.heartBeatSendCode());
    core::Time heartBeatSendInterval = core::Time(0.1);
    socket.setHeartBeatSendInterval(heartBeatSendInterval);
    ASSERT_EQ(heartBeatSendInterval, socket.heartBeatSendInterval());
    bool receiveHeartBeats = true;
    socket.setReceiveHeartBeats(receiveHeartBeats);
    ASSERT_EQ(receiveHeartBeats, socket.receiveHeartBeats());
    NetworkSocket<SerializableDataString, SerializableDataString>::DataStream heartBeatReceiveCode = { 0x1B };
    socket.setHeartBeatReceiveCode(heartBeatReceiveCode);
    ASSERT_EQ(heartBeatReceiveCode, socket.heartBeatReceiveCode());
    core::Time heartBeatTimeout = core::Time(0.123);
    socket.setHeartBeatTimeout(heartBeatTimeout);
    ASSERT_EQ(heartBeatTimeout, socket.heartBeatTimeout());

    // Getters only
    NetworkSocketType type = socket.type();
    ASSERT_EQ(type, NetworkSocketType::UDP);
    std::string name = socket.name();
    ASSERT_EQ(name, "socket");
    std::string remoteIP = socket.remoteIP();
    ASSERT_EQ(remoteIP, "localhost");
    int remotePort = socket.remotePort();
    ASSERT_EQ(remotePort, 5006);
    int ownPort = socket.ownPort();
    ASSERT_EQ(ownPort, 5005);
    unsigned long long maximumMessageSize = socket.maximumMessageSize();
    (void)maximumMessageSize;
    bool isConnected = socket.isConnected();
    ASSERT_EQ(isConnected, false);
    core::Time lastConnectionAttempt = socket.lastConnectionAttempt();
    (void)lastConnectionAttempt;
    core::Time lastHeartBeatSent = socket.lastHeartBeatSent();
    (void)lastHeartBeatSent;
    core::Time lastHeartBeatReceived = socket.lastHeartBeatReceived();
    (void)lastHeartBeatReceived;
}
